"""The WGSI application.
"""
import os
import traceback
import shutil
import io
import mimetypes
import re
import zipfile
import time
import json
import functools
from urllib.parse import urlsplit, urlunsplit, urljoin, quote, unquote
from zlib import adler32
from contextlib import contextmanager
from secrets import token_urlsafe

# dependency
import flask
from flask import request, Response, redirect, abort, render_template
from flask import current_app
from werkzeug.local import LocalProxy
from werkzeug.middleware.proxy_fix import ProxyFix
from werkzeug.datastructures import WWWAuthenticate
from werkzeug.exceptions import HTTPException
from werkzeug.http import is_resource_modified
from werkzeug.http import http_date
from werkzeug.http import parse_options_header, dump_options_header
from werkzeug.utils import cached_property
import jinja2
import commonmark

# this package
from . import *
from . import __version__
from . import Config
from . import util
from .scrapbook import host as wsb_host
from .scrapbook import cache as wsb_cache
from .scrapbook import check as wsb_check
from ._compat.contextlib import nullcontext
from ._compat import zip_stream

# see: https://url.spec.whatwg.org/#percent-encoded-bytes
quote_path = functools.partial(quote, safe=":/[]@!$&'()*+,;=")
quote_path.__doc__ = "Escape reserved chars for the path part of a URL."

bp = flask.Blueprint('default', __name__)
host = LocalProxy(lambda: current_app.config['WEBSCRAPBOOK_HOST'])


def static_url(path):
    return f'{quote_path(request.script_root)}/{quote_path(path)}?a=static'


def static_file(filename, mimetype=None):
    """Output the specified file to the client.

    Args:
        filename: absolute path of the file to output.
    """
    if not os.path.isfile(filename):
        abort(404)
    response = flask.send_file(filename, conditional=True, mimetype=mimetype)
    response.headers.set('Accept-Ranges', 'bytes')
    response.headers.set('Cache-Control', 'no-cache')
    if host.config['app']['content_security_policy'] == 'strict':
        response.headers.set('Content-Security-Policy', "connect-src 'none'; form-action 'none';")
    return response


def zip_static_file(zip, subpath, mimetype=None):
    """Output the specified file in a ZIP to the client.

    Args:
        zip: an opened zipfile.ZipFile
        subpath: str or zipfile.ZipInfo
    """
    if not isinstance(subpath, zipfile.ZipInfo):
        try:
            info = zip.getinfo(subpath)
        except KeyError:
            abort(404)
    else:
        info = subpath

    fh = zip.open(info, 'r')

    lm = util.zip_timestamp(info)
    last_modified = http_date(lm)

    etag = "%s-%s-%s" % (
        lm,
        info.file_size,
        adler32(info.filename.encode("utf-8")) & 0xFFFFFFFF,
        )

    headers = {
        'Accept-Ranges': 'bytes',
        'Cache-Control': 'no-cache',
        'Last-Modified': last_modified,
        'ETag': etag,
        }
    if host.config['app']['content_security_policy'] == 'strict':
        headers['Content-Security-Policy'] = "connect-src 'none'; form-action 'none';"

    response = Response(fh, headers=headers, mimetype=mimetype)
    response.make_conditional(request.environ, accept_ranges=True, complete_length=info.file_size)
    return response


def stream_template(template_name, **context):
    current_app.update_template_context(context)
    t = current_app.jinja_env.get_template(template_name)
    return t.stream(context)


def http_response(body='', status=None, headers=None, format=None):
    """Handle formatted response.

    ref: https://jsonapi.org
    """
    if not format:
        mimetype = None

    # expect body to be a JSON-serializable object
    elif format == 'json':
        mimetype = 'application/json'

        body = {
            'success': True,
            'data': body,
            }

        body = json.dumps(body, ensure_ascii=False)

    # expect body to be a generator of text (mostly JSON) data
    elif format == 'sse':
        mimetype = 'text/event-stream'

        def wrapper(gen):
            try:
                for data in gen:
                    yield "data: " + data + "\n\n"
            except Exception:
                traceback.print_exc()
                err = {'error': {'message': 'Internal Server Error'}}
                yield "data: " + json.dumps(err, ensure_ascii=False) + "\n\n"

            yield "event: complete" + "\n"
            yield "data: " + "\n\n"

        body = wrapper(body)

    else:
        abort(400, f'Output format "{format}" is not supported.')

    return Response(body, status, headers, mimetype=mimetype)


def get_archive_path(filepath):
    """Parse archive file path and the sub-archive path.

    - Priority:
      entry.zip!/entry1.zip!/ = entry.zip!/entry1.zip! >
      entry.zip!/entry1.zip >
      entry.zip!/ = entry.zip! >
      entry.zip

    Returns:
        a list [path-to-directory-or-file]
        or [path-to-zip-file, subpath1, subpath2, ...]
    """
    for m in reversed(list(re.finditer(r'!/', filepath, flags=re.I))):
        archivepath = filepath[:m.start(0)].rstrip('/')
        archivefile = os.path.normpath(os.path.join(host.chroot, archivepath.lstrip('/')))
        conflicting = archivefile + '!'
        if os.path.lexists(conflicting):
            break

        # if parent directory does not exist, FileNotFoundError is raised on
        # Windows, while NotADirectoryError is raised on Linux
        try:
            zip = zipfile.ZipFile(archivefile, 'r')
        except (zipfile.BadZipFile, FileNotFoundError, NotADirectoryError):
            pass
        else:
            with zip as zip:
                def get_subpath(zp, filepath):
                    for m in reversed(list(re.finditer(r'!/', filepath, flags=re.I))):
                        archivepath = filepath[:m.start(0)]
                        conflicting = archivepath + '!/'
                        if any(i.startswith(conflicting) for i in zp.namelist()):
                            break
                        try:
                            with zp.open(archivepath, 'r') as f:
                                f = zip_stream(f)
                                with zipfile.ZipFile(f, 'r') as zip:
                                    rv.append(archivepath)
                                    get_subpath(zip, filepath[m.end(0):])
                                    return
                        except (KeyError, zipfile.BadZipFile):
                            pass
                    rv.append(filepath.rstrip('/'))

                rv = [archivepath]
                get_subpath(zip, filepath[m.end(0):])
                return rv

    return [filepath.rstrip('/')]


@contextmanager
def open_archive_path(paths, mode='r', filters=None):
    """Open the innermost zip.

    Args:
        paths: [path-to-zip-file, subpath1, subpath2, ...]
        mode: 'r' for reading, 'w' for modifying
        filters: a list of file or folder to remove
    """
    last = len(paths) - 1
    if last < 1:
        raise ValueError('length of paths must > 1')

    filtered = False
    stack = []
    try:
        zip = zipfile.ZipFile(paths[0])
        stack.append(zip)
        for i in range(1, last):
            f = zip.open(paths[i])
            f = zip_stream(f)
            stack.append(f)
            zip = zipfile.ZipFile(f)
            stack.append(zip)

        if mode == 'r':
            yield zip

        elif mode == 'w':
            # create a buffer for writing
            buffer = io.BytesIO()
            with zipfile.ZipFile(buffer, 'w') as zip:
                yield zip

            # copy zip file
            for i in reversed(range(1, last + 1)):
                zip0 = stack.pop()
                with zipfile.ZipFile(buffer, 'a') as zip:
                    for info in zip0.infolist():
                        if filters and i == last:
                            if any(info.filename == filter or info.filename.startswith(filter + '/')
                                    for filter in filters):
                                filtered = True
                                continue

                        try:
                            zip.getinfo(info.filename)
                        except KeyError:
                            pass
                        else:
                            continue

                        try:
                            zip.writestr(info, zip0.read(info),
                                    compress_type=info.compress_type,
                                    compresslevel=None if info.compress_type == zipfile.ZIP_STORED else 9)
                        except TypeError:
                            # compresslevel is supported since Python 3.7
                            zip.writestr(info, zip0.read(info),
                                    compress_type=info.compress_type)

                if filters and not filtered:
                    raise KeyError('paths to filter do not exist')

                if i == 1:
                    break

                # writer to another buffer for the parent zip
                buffer2 = io.BytesIO()
                with zipfile.ZipFile(buffer2, 'w') as zip:
                    zip.writestr(paths[i - 1], buffer.getvalue(), compress_type=zipfile.ZIP_STORED)
                buffer.close()
                buffer = buffer2

                # pop a file handler
                stack.pop()

            # write to the outermost zip
            # use 'r+b' as 'wb' causes PermissionError for hidden file in Windows
            buffer.seek(0)
            with open(paths[0], 'r+b') as fw, buffer as fr:
                fw.truncate()
                while True:
                    chunk = fr.read(8192)
                    if not chunk: break
                    fw.write(chunk)
    finally:
        for f in reversed(stack):
            f.close()


def is_local_access():
    """Determine if the client is in same device.
    """
    server_host = request.host.partition(':')[0]
    client_host = request.remote_addr
    return util.is_localhost(server_host) or util.is_localhost(client_host) or server_host == client_host


def get_permission(auth_info, auth_config):
    """Calculate effective permission from provided auth info and config.
    """
    auth = auth_info or {}
    user = auth.get('username') or ''
    pw = auth.get('password') or ''

    for _, entry in auth_config.items():
        entry_user = entry.get('user', '')
        if user != entry_user:
            continue

        entry_pw = entry.get('pw', '')
        entry_pw_salt = entry.get('pw_salt', '')
        entry_pw_type = entry.get('pw_type', '')
        if util.encrypt(pw, entry_pw_salt, entry_pw_type) != entry_pw:
            continue

        entry_permission = entry.get('permission', 'all')
        return entry_permission

    return ''


def verify_authorization(perm, action):
    """Check if authorized or not.
    """
    if perm == 'all':
        return True

    if perm == 'read':
        return action not in {
            'token', 'lock', 'unlock',
            'mkdir', 'mkzip', 'save', 'delete', 'move', 'copy', 'backup',
            'cache', 'check',
            }

    if perm == 'view':
        return action in {'view', 'info', 'source', 'download', 'static'}

    return False


def handle_directory_listing(paths, zip=None, redirect_slash=True, format=None):
    """List contents in a directory.

    Args:
        paths: [path-to-zip-file, subpath1, subpath2, ...]
        zip: an opened zipfile.ZipFile object for faster reading
    """
    # ensure directory has trailing '/'
    if redirect_slash and not request.path.endswith('/'):
        parts = urlsplit(request.url)
        new_url = urlunsplit((
            parts.scheme,
            parts.netloc,
            quote_path(unquote(parts.path)) + '/',
            parts.query,
            parts.fragment,
            ))
        return redirect(new_url)

    # prepare index
    if len(paths) > 1:
        # support 304 if zip not modified
        stats = os.stat(paths[0])
        last_modified = http_date(stats.st_mtime)
        etag = "%s-%s-%s" % (
            stats.st_mtime,
            stats.st_size,
            adler32(paths[0].encode("utf-8")) & 0xFFFFFFFF,
            )

        if not is_resource_modified(request.environ, etag=etag, last_modified=last_modified):
            return http_response(status=304, format=format)

        headers = {
            'Cache-Control': 'no-cache',
            'Last-Modified': last_modified,
            'ETag': etag,
            }

        with nullcontext(zip) if zip else open_archive_path(paths) as zip:
            subentries = util.zip_listdir(zip, paths[-1])

    else:
        # disallow cache to reflect any content file change
        stats = os.stat(paths[0])
        headers = {
            'Cache-Control': 'no-store',
            'Last-Modified': http_date(stats.st_mtime),
            }

        subentries = util.listdir(paths[0])

    if format == 'sse':
        def gen():
            for entry in subentries:
                data = {
                    'name': entry.name,
                    'type': entry.type,
                    'size': entry.size,
                    'last_modified': entry.last_modified,
                    }

                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), headers=headers, format=format)

    if format == 'json':
        data = []
        for entry in subentries:
            data.append({
                    'name': entry.name,
                    'type': entry.type,
                    'size': entry.size,
                    'last_modified': entry.last_modified,
                    })
        return http_response(data, headers=headers, format=format)

    body = render_template('index.html',
            sitename=host.name,
            is_local=is_local_access(),
            base=request.script_root,
            path=request.path,
            pathparts=request.paths,
            subentries=subentries,
            )
    return http_response(body, headers=headers)


def handle_archive_viewing(paths, mimetype):
    """Handle direct visit of HTZ/MAFF file.

    Args:
        paths: [path-to-zip-file, subpath1, subpath2, ...]
    """
    def list_maff_pages(pages):
        """List available web pages in a MAFF file.
        """
        return render_template('maff_index.html',
                sitename=host.name,
                is_local=is_local_access(),
                base=request.script_root,
                path=request.path,
                pages=pages,
                )

    if mimetype == "application/html+zip":
        subpath = "index.html"
    else:
        if len(paths) > 1:
            with open_archive_path(paths) as zip:
                with zip.open(paths[-1]) as zh:
                    pages = util.get_maff_pages(zh)
        else:
            pages = util.get_maff_pages(paths[-1])

        if len(pages) > 1:
            # multiple index files
            return list_maff_pages(pages)

        if len(pages) == 0:
            # no valid index file found
            return list_maff_pages([])

        subpath = pages[0].indexfilename

    parts = urlsplit(request.url)
    new_url = urlunsplit((
        parts.scheme,
        parts.netloc,
        quote_path(unquote(parts.path)) + '!/' + quote_path(subpath),
        parts.query,
        parts.fragment,
        ))
    return redirect(new_url)


def handle_markdown_output(paths, zip=None):
    """Output processed markdown.

    Args:
        paths: [path-to-zip-file, subpath1, subpath2, ...]
        zip: an opened zipfile.ZipFile object for faster reading
    """
    if len(paths) > 1:
        if zip:
            context = nullcontext(zip)
        else:
            context = open_archive_path(paths)
    else:
        context = nullcontext(None)

    with context as zip:
        # calculate last-modified time and etag
        if zip:
            info = zip.getinfo(paths[-1])
            lm = util.zip_timestamp(info)
            last_modified = http_date(lm)

            etag = "%s-%s-%s" % (
                lm,
                info.file_size,
                adler32(info.filename.encode("utf-8")) & 0xFFFFFFFF,
                )
        else:
            stats = os.stat(paths[0])
            last_modified = http_date(stats.st_mtime)
            etag = "%s-%s-%s" % (
                stats.st_mtime,
                stats.st_size,
                adler32(paths[0].encode("utf-8")) & 0xFFFFFFFF,
                )

        if not is_resource_modified(request.environ, etag=etag, last_modified=last_modified):
            return http_response(status=304)

        headers = {
            'Cache-Control': 'no-cache',
            'Last-Modified': last_modified,
            'ETag': etag,
            }
        if host.config['app']['content_security_policy'] == 'strict':
            headers['Content-Security-Policy'] = "connect-src 'none'; form-action 'none';"

        # prepare content
        if zip:
            with zip.open(info) as f:
                body = f.read().decode('UTF-8')
        else:
            with open(paths[0], 'r', encoding='UTF-8') as f:
                body = f.read()

    body = render_template('markdown.html',
            sitename=host.name,
            is_local=is_local_access(),
            base=request.script_root,
            path=request.path,
            pathparts=request.paths,
            content=commonmark.commonmark(body),
            )

    return http_response(body, headers=headers)


class Request(flask.Request):
    """Subclassed Request object for more useful properties.
    """
    @cached_property
    def paths(self):
        """Like request.path, but with ZIP subpaths resolved."""
        return get_archive_path(self.path)

    @cached_property
    def localpath(self):
        """Corresponding filesystem path of the requested path."""
        return os.path.normpath(os.path.join(host.chroot, self.path.strip('/')))

    @cached_property
    def localpaths(self):
        """Like localpath, but with ZIP subpaths resolved."""
        paths = self.paths.copy()
        paths[0] = os.path.normpath(os.path.join(host.chroot, paths[0].lstrip('/')))
        return paths

    @cached_property
    def localrealpath(self):
        """Like localpath, but with symlinks resolved."""
        return os.path.realpath(self.localpath)

    @cached_property
    def localmimetype(self):
        """Mimetype of the requested path."""
        mimetype, _ = mimetypes.guess_type(self.localrealpath)
        return mimetype

    @cached_property
    def action(self):
        """Shortcut of the requested action."""
        rv = request.values.get('a', default='view')
        rv = request.values.get('action', default=rv)
        return rv

    @cached_property
    def format(self):
        """Shortcut of the requested format."""
        rv = request.values.get('f')
        rv = request.values.get('format', default=rv)
        return rv


def handle_action_token(func):
    """A decorator function that validates token.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        token = request.values.get('token') or ''

        if not host.token_validate(token):
            abort(400, 'Invalid access token.')

        host.token_delete(token)

        return func(*args, **kwargs)

    return wrapper


def handle_action_advanced(func):
    """A decorator function that helps handling an advanced command.

    - Verify POST method.
    - Provide a default return value.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        format = request.format

        if request.method != 'POST':
            abort(405, valid_methods=['POST'])

        rv = func(*args, **kwargs)

        if rv is not None:
            return rv

        if format:
            return http_response('Command run successfully.', format=format)

        return http_response(status=204)

    return wrapper


def handle_action_writing(func):
    """A decorator function that helps handling a writing action.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if os.path.abspath(request.localpath) == host.chroot:
            abort(403, "Unable to operate the root directory.")

        return func(*args, **kwargs)

    return wrapper


def handle_action_renaming(func):
    """A decorator function that helps handling a move/copy action.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        localpaths = request.localpaths

        if len(localpaths) > 1:
            with open_archive_path(localpaths) as zip:
                try:
                    zip.getinfo(localpaths[-1])
                except KeyError:
                    if not util.zip_hasdir(zip, localpaths[-1] + '/'):
                        abort(404, "Source does not exist.")
        else:
            if not os.path.lexists(localpaths[0]):
                abort(404, "Source does not exist.")

        target = request.values.get('target')

        if target is None:
            abort(400, 'Target is not specified.')

        targetpaths = get_archive_path(target)
        targetpaths[0] = os.path.normpath(os.path.join(host.chroot, targetpaths[0].lstrip('/')))

        if not targetpaths[0].startswith(os.path.join(host.chroot, '')):
            abort(403, "Unable to operate beyond the root directory.")

        if len(targetpaths) > 1:
            with open_archive_path(targetpaths) as zip:
                try:
                    zip.getinfo(targetpaths[-1])
                except KeyError:
                    if util.zip_hasdir(zip, targetpaths[-1] + '/'):
                        abort(400, 'Found something at target.')
                else:
                    abort(400, 'Found something at target.')
        else:
            if os.path.lexists(targetpaths[0]):
                abort(400, 'Found something at target.')

        return func(sourcepaths=localpaths, targetpaths=targetpaths, *args, **kwargs)

    return wrapper


def action_unknown():
    """Default handler for an undefined action"""
    abort(400, "Action not supported.")


def action_view():
    """Show the content of a file or list a directory.

    If formatted, show information of the file or directory.
    """
    # info for other output formats
    if request.format:
        return action_info()

    localpaths = request.localpaths
    mimetype = request.localmimetype

    if len(localpaths) > 1:
        with open_archive_path(localpaths) as zip:
            try:
                info = zip.getinfo(localpaths[-1])
            except KeyError:
                # File does not exist.  List directory only when URL
                # suffixed with "/", as it's not a common operation,
                # and it's costy to check for directory existence in
                # a ZIP.
                if request.path.endswith('/'):
                    try:
                        return handle_directory_listing(localpaths, zip, redirect_slash=False)
                    except util.ZipDirNotFoundError:
                        abort(404)
                abort(404)
            else:
                # view archive file
                if mimetype in ("application/html+zip", "application/x-maff"):
                    return handle_archive_viewing(localpaths, mimetype)

                # view markdown
                if mimetype == "text/markdown":
                    return handle_markdown_output(localpaths, zip)

                # convert meta refresh to 302 redirect
                if localpaths[-1].lower().endswith('.htm'):
                    with zip.open(info) as fh:
                        target = util.parse_meta_refresh(fh).target

                    if target is not None:
                        # Keep several chars as javascript encodeURI do,
                        # plus "%" as target may have already been escaped.
                        parts = urlsplit(urljoin(request.url, quote(target, ";,/?:@&=+$-_.!~*'()#%")))
                        new_url = urlunsplit((
                            parts.scheme,
                            parts.netloc,
                            quote_path(unquote(parts.path)),
                            parts.query,
                            parts.fragment,
                            ))
                        return redirect(new_url)

                # show static file for other cases
                response = zip_static_file(zip, localpaths[-1], mimetype=mimetype)
    else:
        localpath = localpaths[0]

        # handle directory
        if os.path.isdir(localpath):
            return handle_directory_listing(localpaths)

        # handle file
        elif os.path.isfile(localpath):
            # view archive file
            if mimetype in ("application/html+zip", "application/x-maff"):
                return handle_archive_viewing(localpaths, mimetype)

            # view markdown
            if mimetype == "text/markdown":
                return handle_markdown_output(localpaths)

            # convert meta refresh to 302 redirect
            if request.localrealpath.lower().endswith('.htm'):
                target = util.parse_meta_refresh(localpath).target

                if target is not None:
                    # Keep several chars as javascript encodeURI do,
                    # plus "%" as target may have already been escaped.
                    parts = urlsplit(urljoin(request.url, quote(target, ";,/?:@&=+$-_.!~*'()#%")))
                    new_url = urlunsplit((
                        parts.scheme,
                        parts.netloc,
                        quote_path(unquote(parts.path)),
                        parts.query,
                        parts.fragment,
                        ))
                    return redirect(new_url)

            # show static file for other cases
            response = static_file(localpath, mimetype=mimetype)

        else:
            abort(404)

    # don't include charset
    m, p = parse_options_header(response.headers.get('Content-Type'))
    try:
        del p['charset']
    except KeyError:
        pass
    response.headers.set('Content-Type', dump_options_header(m, p))

    return response


def action_source():
    """Show file content as plain text."""
    if request.format:
        abort(400, "Action not supported.")

    localpaths = request.localpaths

    if len(localpaths) > 1:
        with open_archive_path(localpaths) as zip:
            response = zip_static_file(zip, localpaths[-1])
    else:
        response = static_file(localpaths[0])

    # show as inline plain text
    # @TODO: Chromium (80) seems to ignore header mimetype for certain types
    #        like image and zip
    encoding = request.values.get('e', 'utf-8')
    encoding = request.values.get('encoding', default=encoding)
    response.headers.set('Content-Type', 'text/plain; charset=' + quote(encoding))
    response.headers.set('Content-Disposition', 'inline')

    return response


def action_download():
    """Download the  file."""
    if request.format:
        abort(400, "Action not supported.")

    localpaths = request.localpaths

    if len(localpaths) > 1:
        with open_archive_path(localpaths) as zip:
            response = zip_static_file(zip, localpaths[-1], mimetype=request.localmimetype)
    else:
        response = static_file(localpaths[0])

    filename = quote_path(os.path.basename(request.localrealpath))
    response.headers.set('Content-Disposition',
            f'''attachment; filename*=UTF-8''{filename}; filename="{filename}"''')
    return response


def action_info():
    """Show information of a path."""
    format = request.format

    if not format:
        abort(400, "Action not supported.")

    localpaths = request.localpaths
    mimetype = request.localmimetype

    if len(localpaths) > 1:
        with open_archive_path(localpaths) as zip:
            info = util.zip_file_info(zip, localpaths[-1])
    else:
        info = util.file_info(localpaths[0])

    data = {
        'name': info.name,
        'type': info.type,
        'size': info.size,
        'last_modified': info.last_modified,
        'mime': mimetype,
        }
    return http_response(data, format=format)


def action_list():
    """List entries in a directory."""
    format = request.format

    if not format:
        abort(400, "Action not supported.")

    localpaths = request.localpaths

    if len(localpaths) > 1:
        try:
            return handle_directory_listing(localpaths, redirect_slash=False, format=format)
        except util.ZipDirNotFoundError:
            abort(404, "Directory does not exist.")

    if os.path.isdir(localpaths[0]):
        return handle_directory_listing(localpaths, redirect_slash=False, format=format)

    abort(404, "Directory does not exist.")


def action_static():
    """Show a static file of the current theme."""
    format = request.format

    if format:
        abort(400, "Action not supported.")

    filepath = request.path.strip('/')
    file = host.get_static_file(filepath)
    if file:
        return static_file(file)

    abort(404)


def action_edit():
    """Simple text editor for a file."""
    format = request.format

    if format:
        abort(400, "Action not supported.")

    localpaths = request.localpaths
    localpath = localpaths[0]

    if os.path.lexists(localpath) and not os.path.isfile(localpath):
        abort(400, "Found a non-file here.")

    if len(localpaths) > 1:
        with open_archive_path(localpaths) as zip:
            try:
                info = zip.getinfo(localpaths[-1])
            except KeyError:
                body = b''
            else:
                body = zip.read(info)
    else:
        try:
            with open(localpath, 'rb') as f:
                body = f.read()
        except FileNotFoundError:
            body = b''

    encoding = request.values.get('e')
    encoding = request.values.get('encoding', default=encoding)

    try:
        body = body.decode(encoding or 'UTF-8')
    except (LookupError, UnicodeDecodeError):
        encoding = 'ISO-8859-1'
        body = body.decode(encoding)

    body = render_template('edit.html',
            sitename=host.name,
            is_local=is_local_access(),
            base=request.script_root,
            path=request.path,
            body=body,
            encoding=encoding,
            )

    return http_response(body, format=format)


def action_editx():
    """HTML editor for a file."""
    format = request.format

    if format:
        abort(400, "Action not supported.")

    localpaths = request.localpaths
    localpath = localpaths[0]

    if os.path.lexists(localpath) and not os.path.isfile(localpath):
        abort(400, "Found a non-file here.")

    if not request.localmimetype in ("text/html", "application/xhtml+xml"):
        abort(400, "This is not an HTML file.")

    if len(localpaths) > 1:
        with open_archive_path(localpaths) as zip:
            try:
                info = zip.getinfo(localpaths[-1])
            except KeyError:
                abort(404)
    else:
        if not os.path.lexists(localpath):
            abort(404)

    body = render_template('editx.html',
            sitename=host.name,
            is_local=is_local_access(),
            base=request.script_root,
            path=request.path,
            )

    return http_response(body, format=format)


def action_exec():
    """Launch a file or directory."""
    format = request.format

    if not is_local_access():
        abort(400, "Command can only run on local device.")

    localpath = request.localpath

    if not os.path.lexists(localpath):
        abort(404, "File does not exist.")

    util.launch(localpath)

    if format:
        return http_response('Command run successfully.', format=format)

    return http_response(status=204)


def action_browse():
    """Open a file or directory in the file browser."""
    format = request.format

    if not is_local_access():
        abort(400, "Command can only run on local device.")

    localpath = request.localpath

    if not os.path.lexists(localpath):
        abort(404, "File does not exist.")

    util.view_in_explorer(localpath)

    if format:
        return http_response('Command run successfully.', format=format)

    return http_response(status=204)


def action_config():
    """Show server config."""
    format = request.format

    if not format:
        abort(400, "Action not supported.")

    data = host.config.dump_object()

    # filter values for better security
    data = {k:v for k, v in data.items() if k in ('app', 'book')}
    data['app'] = {k:v for k, v in data['app'].items() if k in ('name', 'theme')}

    # add and rewrite values for client to better know the server
    data['app']['base'] = request.script_root
    data['app']['is_local'] = is_local_access()
    data['VERSION'] = __version__
    data['WSB_DIR'] = WSB_DIR
    data['WSB_CONFIG'] = WSB_CONFIG
    data['WSB_EXTENSION_MIN_VERSION'] = WSB_EXTENSION_MIN_VERSION

    return http_response(data, format=format)


def action_token():
    """Acquire a token and return its name."""
    format = request.format

    # require POST method
    if request.method != 'POST':
        abort(405, valid_methods=['POST'])

    return http_response(host.token_acquire(), format=format)


@handle_action_advanced
@handle_action_token
def action_lock():
    """Acquire a lock for the given name.

    URL params:
    - id: for persisting and extending previous lock.
    - chkt: timeout to wait for the lock.
    """
    # verify name
    name = request.values.get('name')
    if name is None:
        abort(400, "Lock name is not specified.")

    id = request.values.get('id')

    timeout = request.values.get('chkt', 5, type=float)

    try:
        lock = host.get_lock(name, persist=id)
    except wsb_host.LockPersistError:
        abort(400, f'Unable to persist lock "{name}".')

    if id:
        try:
            lock.extend()
        except wsb_host.LockExtendNotFoundError:
            # Lock file gone in this short interval. Try acquire.
            pass
        except wsb_host.LockExtendError:
            abort(500, f'Unable to extend lock "{name}".')
        else:
            return http_response(lock.id, format=request.format)

    try:
        lock.acquire(timeout=timeout)
    except wsb_host.LockTimeoutError:
        abort(503, f'Unable to acquire lock "{name}".', retry_after=60)
    except wsb_host.LockRegenerateError:
        abort(500, f'Unable to regenerate stale lock "{name}".')
    except wsb_host.LockGenerateError:
        abort(500, f'Unable to create lock "{name}".')

    return http_response(lock.id, format=request.format)


@handle_action_advanced
@handle_action_token
def action_unlock():
    """Release a lock for the given name."""
    # verify name
    name = request.values.get('name')
    if name is None:
        abort(400, "Lock name is not specified.")

    # verify ID
    id = request.values.get('id')
    if id is None:
        abort(400, "Lock ID is not specified.")

    try:
        lock = host.get_lock(name, persist=id)
    except wsb_host.LockPersistError:
        abort(400, f'Unable to persist lock "{name}".')

    try:
        lock.release()
    except wsb_host.LockReleaseNotFoundError:
        pass
    except wsb_host.LockReleaseError:
        abort(500, f'Unable to remove lock "{name}".')


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_mkdir():
    """Create a directory."""
    format = request.format
    localpaths = request.localpaths

    if len(localpaths) > 1:
        try:
            folderpath = localpaths[-1] + '/'
            zip = None

            with open_archive_path(localpaths) as zip0:
                try:
                    zip0.getinfo(folderpath)
                except KeyError:
                    # append for a non-nested zip
                    if len(localpaths) == 2:
                        zip = zipfile.ZipFile(localpaths[0], 'a')
                else:
                    # skip as the folder already exists
                    return

            if zip is None:
                zip = open_archive_path(localpaths, 'w')

            with zip as zip:
                info = zipfile.ZipInfo(folderpath, time.localtime())
                zip.writestr(info, b'', compress_type=zipfile.ZIP_STORED)
        except Exception:
            traceback.print_exc()
            abort(500, "Unable to write to this ZIP file.")

    else:
        localpath = localpaths[0]

        if os.path.lexists(localpath) and not os.path.isdir(localpath):
            abort(400, "Found a non-directory here.")

        try:
            os.makedirs(localpath, exist_ok=True)
        except OSError:
            traceback.print_exc()
            abort(500, "Unable to create a directory here.")


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_mkzip():
    """Create a zip file."""
    format = request.format
    localpaths = request.localpaths

    if len(localpaths) > 1:
        try:
            zip = None

            # append for a nonexistent path in a non-nested zip
            if len(localpaths) == 2:
                zip0 = zipfile.ZipFile(localpaths[0], 'a')
                try:
                    zip0.getinfo(localpaths[-1])
                except KeyError:
                    zip = zip0
                except:
                    zip0.close()
                    raise
                else:
                    zip0.close()

            if zip is None:
                zip = open_archive_path(localpaths, 'w')

            with zip as zip:
                info = zipfile.ZipInfo(localpaths[-1], time.localtime())
                buf = io.BytesIO()
                with zipfile.ZipFile(buf, 'w'):
                    pass
                zip.writestr(info, buf.getvalue(), compress_type=zipfile.ZIP_STORED)
        except Exception:
            traceback.print_exc()
            abort(500, "Unable to write to this ZIP file.")

    else:
        localpath = localpaths[0]

        if os.path.lexists(localpath) and not os.path.isfile(localpath):
            abort(400, "Found a non-file here.")

        try:
            os.makedirs(os.path.dirname(localpath), exist_ok=True)
        except Exception:
            traceback.print_exc()
            abort(500, "Unable to write to this path.")

        try:
            with zipfile.ZipFile(localpath, 'w') as f:
                pass
        except Exception:
            traceback.print_exc()
            abort(500, "Unable to write to this file.")


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_save():
    """Write a file with provided text or uploaded stream."""
    format = request.format
    localpaths = request.localpaths

    if len(localpaths) > 1:
        try:
            zip = None

            # append for a nonexistent path in a non-nested zip
            if len(localpaths) == 2:
                zip0 = zipfile.ZipFile(localpaths[0], 'a')
                try:
                    zip0.getinfo(localpaths[-1])
                except KeyError:
                    zip = zip0
                except Exception:
                    zip0.close()
                    raise
                else:
                    zip0.close()

            if zip is None:
                zip = open_archive_path(localpaths, 'w')

            with zip as zip:
                info = zipfile.ZipInfo(localpaths[-1], time.localtime())
                file = request.files.get('upload')
                if file is not None:
                    with zip.open(info, 'w', force_zip64=True) as fh:
                        stream = file.stream
                        while True:
                            chunk = stream.read(8192)
                            if not chunk: break
                            fh.write(chunk)
                else:
                    bytes_ = request.values.get('text', '').encode('ISO-8859-1')
                    try:
                        zip.writestr(info, bytes_, compress_type=zipfile.ZIP_DEFLATED, compresslevel=9)
                    except TypeError:
                        # compresslevel is supported since Python 3.7
                        zip.writestr(info, bytes_, compress_type=zipfile.ZIP_DEFLATED)
        except Exception:
            traceback.print_exc()
            abort(500, "Unable to write to this ZIP file.")

    else:
        localpath = localpaths[0]

        if os.path.lexists(localpath) and not os.path.isfile(localpath):
            abort(400, "Found a non-file here.")

        try:
            os.makedirs(os.path.dirname(localpath), exist_ok=True)
        except OSError:
            traceback.print_exc()
            abort(500, "Unable to write to this path.")

        try:
            file = request.files.get('upload')
            if file is not None:
                file.save(localpath)
            else:
                bytes_ = request.values.get('text', '').encode('ISO-8859-1')
                with open(localpath, 'wb') as f:
                    f.write(bytes_)
        except Exception:
            traceback.print_exc()
            abort(500, "Unable to write to this file.")


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_delete():
    """Delete a file or directory."""
    format = request.format
    localpaths = request.localpaths

    if len(localpaths) > 1:
        try:
            with open_archive_path(localpaths, 'w', [localpaths[-1]]) as zip:
                pass
        except KeyError:
            # fail since nothing is deleted
            abort(404, "Entry does not exist in this ZIP file.")
        except Exception:
            traceback.print_exc()
            abort(500, "Unable to write to this ZIP file.")

    else:
        localpath = localpaths[0]

        if not os.path.lexists(localpath):
            abort(404, "File does not exist.")

        if util.file_is_link(localpath):
            try:
                os.remove(localpath)
            except OSError:
                traceback.print_exc()
                abort(500, "Unable to delete this link.")
        elif os.path.isfile(localpath):
            try:
                os.remove(localpath)
            except OSError:
                traceback.print_exc()
                abort(500, "Unable to delete this file.")
        elif os.path.isdir(localpath):
            try:
                shutil.rmtree(localpath)
            except OSError:
                traceback.print_exc()
                abort(500, "Unable to delete this directory.")
        else:
            # this should not happen
            abort(500, "Unable to handle this path.")


@handle_action_advanced
@handle_action_token
@handle_action_writing
@handle_action_renaming
def action_move(sourcepaths, targetpaths):
    """Move a file or directory."""
    format = request.format

    try:
        if len(sourcepaths) == 1:
            if len(targetpaths) == 1:
                try:
                    os.makedirs(os.path.dirname(targetpaths[0]), exist_ok=True)
                except OSError:
                    traceback.print_exc()
                    abort(500, "Unable to copy to this path.")

                shutil.move(sourcepaths[0], targetpaths[0])

            else:
                # Moving a file into a zip is like moving across disk,
                # which makes little sense. Additionally, moving a
                # symlink/junction should rename the entry and cannot be
                # implemented as copying-deleting. Forbid such operation to
                # prevent a confusion.
                abort(400, "Unable to move across a zip.")

        elif len(sourcepaths) > 1:
            if len(targetpaths) == 1:
                # Moving from zip to disk is like moving across disk, which
                # makes little sense.
                abort(400, "Unable to move across a zip.")

            else:
                with open_archive_path(sourcepaths) as zip:
                    try:
                        zip.getinfo(sourcepaths[-1])
                    except KeyError:
                        entries = [e for e in zip.namelist() if e.startswith(sourcepaths[-1] + '/')]
                    else:
                        entries = [sourcepaths[-1]]

                    with open_archive_path(targetpaths, 'w') as zip2:
                        cut = len(sourcepaths[-1])
                        for entry in entries:
                            info = zip.getinfo(entry)
                            info.filename = targetpaths[-1] + entry[cut:]
                            try:
                                zip2.writestr(info, zip.read(entry),
                                        compresslevel=None if info.compress_type == zipfile.ZIP_STORED else 9)
                            except TypeError:
                                # compresslevel is supported since Python 3.7
                                zip2.writestr(info, zip.read(entry))

                with open_archive_path(sourcepaths, 'w', entries) as zip:
                    pass

    except HTTPException:
        raise
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to move to the target.')


@handle_action_advanced
@handle_action_token
@handle_action_writing
@handle_action_renaming
def action_copy(sourcepaths, targetpaths):
    """Copy a file or directory."""
    format = request.format

    # Copying a symlink/junction means copying the real file/directory.
    # It makes no sense if the symlink/junction is broken.
    if not os.path.exists(sourcepaths[0]):
        abort(404, "Source does not exist.")

    try:
        if len(sourcepaths) == 1:
            if len(targetpaths) == 1:
                try:
                    os.makedirs(os.path.dirname(targetpaths[0]), exist_ok=True)
                except OSError:
                    traceback.print_exc()
                    abort(500, "Unable to copy to this path.")

                try:
                    shutil.copytree(sourcepaths[0], targetpaths[0])
                except NotADirectoryError:
                    shutil.copy2(sourcepaths[0], targetpaths[0])
                except shutil.Error:
                    traceback.print_exc()
                    abort(500, 'Fail to copy some files.')

            else:
                error = False
                with open_archive_path(targetpaths, 'w') as zip:
                    try:
                        util.zip_compress(zip, sourcepaths[0], targetpaths[-1])
                    except shutil.Error:
                        traceback.print_exc()
                        error = True
                if error:
                    abort(500, 'Fail to copy some files.')

        elif len(sourcepaths) > 1:
            if len(targetpaths) == 1:
                try:
                    os.makedirs(os.path.dirname(targetpaths[0]), exist_ok=True)
                except OSError:
                    traceback.print_exc()
                    abort(500, "Unable to copy to this path.")

                with open_archive_path(sourcepaths) as zh:
                    util.zip_extract(zh, targetpaths[0], sourcepaths[-1])

            else:
                with open_archive_path(sourcepaths) as zip:
                    try:
                        zip.getinfo(sourcepaths[-1])
                    except KeyError:
                        entries = [e for e in zip.namelist() if e.startswith(sourcepaths[-1] + '/')]
                    else:
                        entries = [sourcepaths[-1]]

                    with open_archive_path(targetpaths, 'w') as zip2:
                        cut = len(sourcepaths[-1])
                        for entry in entries:
                            info = zip.getinfo(entry)
                            info.filename = targetpaths[-1] + entry[cut:]
                            try:
                                zip2.writestr(info, zip.read(entry),
                                        compresslevel=None if info.compress_type == zipfile.ZIP_STORED else 9)
                            except TypeError:
                                # compresslevel is supported since Python 3.7
                                zip2.writestr(info, zip.read(entry))

    except HTTPException:
        raise
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to copy to the target.')


@handle_action_advanced
@handle_action_token
def action_backup():
    """Bakup file or directory."""
    format = request.format
    localpaths = request.localpaths

    if len(localpaths) > 1:
        abort(400, "Unable to backup inside a zip file.")

    ts = request.values.get('ts') or util.datetime_to_id()
    move = request.values.get('move', default=False, type=bool)

    host.backup(localpaths[0], os.path.join(host.backup_dir, ts), move=move)


@handle_action_token
def action_cache():
    """Invoke the cacher."""
    format = request.format

    kwargs = {
        'book_ids': request.values.getlist('book'),
        'item_ids': request.values.getlist('item'),
        'no_lock': request.values.get('no_lock', default=False, type=bool),
        'no_backup': request.values.get('no_backup', default=False, type=bool),
        'fulltext': request.values.get('fulltext', default=False, type=bool),
        'inclusive_frames': request.values.get('inclusive_frames', default=False, type=bool),
        'recreate': request.values.get('recreate', default=False, type=bool),
        'static_site': request.values.get('static_site', default=False, type=bool),
        'static_index': request.values.get('static_index', default=False, type=bool),
        'rss_root': request.values.get('rss_root'),
        'rss_item_count': request.values.get('rss_item_count', default=50, type=int),
        'locale': request.values.get('locale'),
        }

    headers = {
        'Cache-Control': 'no-store',
        }
    root = host.root
    config = host.config

    if format == 'sse':
        def gen():
            for info in wsb_cache.generate(root, config=config, **kwargs):
                data = {
                    'type': info.type,
                    'msg': info.msg,
                    }

                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), headers=headers, format=format)

    elif format:
        abort(400, "Action not supported.")

    def gen():
        yield from wsb_cache.generate(root, config=config, **kwargs)

    stream = stream_template('cli.html',
        title=f'Indexing...',
        messages=gen(),
        debug=False,
        )

    return Response(stream, headers=headers)


@handle_action_token
def action_check():
    """Invoke the checker."""
    format = request.format

    kwargs = {
        'book_ids': request.values.getlist('book'),
        'no_lock': request.values.get('no_lock', default=False, type=bool),
        'no_backup': request.values.get('no_backup', default=False, type=bool),
        'resolve_invalid_id': request.values.get('resolve_invalid_id', default=False, type=bool),
        'resolve_missing_index': request.values.get('resolve_missing_index', default=False, type=bool),
        'resolve_missing_index_file': request.values.get('resolve_missing_index_file', default=False, type=bool),
        'resolve_missing_date': request.values.get('resolve_missing_date', default=False, type=bool),
        'resolve_older_mtime': request.values.get('resolve_older_mtime', default=False, type=bool),
        'resolve_toc_unreachable': request.values.get('resolve_toc_unreachable', default=False, type=bool),
        'resolve_toc_invalid': request.values.get('resolve_toc_invalid', default=False, type=bool),
        'resolve_toc_empty_subtree': request.values.get('resolve_toc_empty_subtree', default=False, type=bool),
        'resolve_unindexed_files': request.values.get('resolve_unindexed_files', default=False, type=bool),
        'resolve_absolute_icon': request.values.get('resolve_absolute_icon', default=False, type=bool),
        'resolve_unused_icon': request.values.get('resolve_unused_icon', default=False, type=bool),
        }

    headers = {
        'Cache-Control': 'no-store',
        }
    root = host.root
    config = host.config

    if format == 'sse':
        def gen():
            for info in wsb_check.run(root, config=config, **kwargs):
                data = {
                    'type': info.type,
                    'msg': info.msg,
                    }

                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), headers=headers, format=format)

    elif format:
        abort(400, "Action not supported.")

    def gen():
        yield from wsb_check.run(root, config=config, **kwargs)

    stream = stream_template('cli.html',
        title=f'Indexing...',
        messages=gen(),
        debug=False,
        )

    return Response(stream, headers=headers)


@bp.before_request
def handle_before_request():
    # replace SCRIPT_NAME with the custom if set
    if host.config['app']['base']:
        # Flask treats SCRIPT_NAME in the same way as PATH_INFO, which is an
        # IRI string decoded as ISO-8859-1 according to WSGI standard).
        request.environ['SCRIPT_NAME'] = unquote(host.config['app']['base']).encode('UTF-8').decode('ISO-8859-1')

    # handle authorization
    try:
        auth_config = host.config['auth']
    except KeyError:
        # auth not required
        return

    perm = get_permission(request.authorization, auth_config)
    if not verify_authorization(perm, request.action):
        auth = WWWAuthenticate()
        auth.set_basic(host.config['app']['name'])
        abort(401, 'You are not authorized.', www_authenticate=auth)


@bp.route('/', methods=['GET', 'HEAD', 'POST'])
@bp.route('/<path:filepath>', methods=['GET', 'HEAD', 'POST'])
def handle_request(filepath=''):
    """Handle an HTTP request (HEAD, GET, POST).
    """
    try:
        handler = globals().get(f'action_{request.action}') or action_unknown
        return handler()
    except PermissionError:
        abort(403)


@bp.after_request
def handle_after_request(response):
    # forbid a privileged page to be framed
    if host.config['app']['content_security_policy'] == 'strict':
        if 'Content-Security-Policy' not in response.headers:
            response.headers.set('Content-Security-Policy', "frame-ancestors 'none';")
            response.headers.set('X-Frame-Options', 'deny')

    return response


@bp.errorhandler(HTTPException)
def handle_error(exc):
    """Handle formatted error if requested by client.
    """
    if request.format == 'json':
        response = exc.get_response()
        response.data = json.dumps({
            'error': {
                'status': exc.code,
                'message': exc.description,
                },
            })
        response.content_type = "application/json"
        return response

    return exc


class WebHost(wsb_host.Host):
    """Extended Host class that also handles HTTP server related things.

    - Token handling: security token validation to avoid CSRF attack.
    """
    TOKEN_PURGE_INTERVAL = 3600  # in seconds
    TOKEN_DEFAULT_EXPIRY = 1800  # in seconds

    def __init__(self, root, config=None):
        super().__init__(root, config=config)

        # token handling
        self.tokens = os.path.join(self.root, WSB_DIR, 'server', 'tokens')
        self.token_last_purge = 0

    def token_acquire(self, now=None):
        if now is None:
            now = int(time.time())

        self.token_check_delete_expire(now)

        token = token_urlsafe()
        token_file = os.path.join(self.tokens, token)
        while os.path.lexists(token_file):
            token = token_urlsafe()
            token_file = os.path.join(self.tokens, token)

        os.makedirs(os.path.dirname(token_file), exist_ok=True)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(now + self.TOKEN_DEFAULT_EXPIRY))

        return token

    def token_validate(self, token, now=None):
        if now is None:
            now = int(time.time())

        token_file = os.path.join(self.tokens, token)

        try:
            with open(token_file, 'r', encoding='UTF-8') as f:
                expire = int(f.read())
        except (FileNotFoundError, IsADirectoryError):
            return False

        if now >= expire:
            os.remove(token_file)
            return False

        return True

    def token_delete(self, token):
        token_file = os.path.join(self.tokens, token)

        try:
            os.remove(token_file)
        except OSError:
            pass

    def token_delete_expire(self, now=None):
        if now is None:
            now = int(time.time())

        try:
            token_files = os.scandir(self.tokens)
        except FileNotFoundError:
            pass
        else:
            for token_file in token_files:
                try:
                    with open(token_file, 'r', encoding='UTF-8') as f:
                        expire = int(f.read())
                except (OSError, ValueError):
                    continue
                if now >= expire:
                    os.remove(token_file)

    def token_check_delete_expire(self, now=None):
        if now is None:
            now = int(time.time())

        if now >= self.token_last_purge + self.TOKEN_PURGE_INTERVAL:
            self.token_last_purge = now
            self.token_delete_expire(now)


def make_app(root=".", config=None):
    _host = WebHost(root, config=config)

    # main app instance
    app = flask.Flask(__name__, instance_path=_host.chroot)
    app.register_blueprint(bp)
    app.request_class = Request
    app.config['WEBSCRAPBOOK_HOST'] = _host

    xheaders = {
            'x_for': _host.config['app']['allowed_x_for'],
            'x_proto': _host.config['app']['allowed_x_proto'],
            'x_host': _host.config['app']['allowed_x_host'],
            'x_port': _host.config['app']['allowed_x_port'],
            'x_prefix': _host.config['app']['allowed_x_prefix'],
            }

    if any(v for v in xheaders.values()):
        app.wsgi_app = ProxyFix(app.wsgi_app, **xheaders)

    app.jinja_loader = jinja2.FileSystemLoader(_host.templates)
    app.jinja_env.globals.update({
            'os': os,
            'time': time,
            'get_breadcrumbs': util.get_breadcrumbs,
            'format_filesize': util.format_filesize,
            'quote_path': quote_path,
            'static_url': static_url,
            })

    return app



from unittest import mock
import unittest
import os
import shutil
import glob
import time

from lxml import etree

from webscrapbook import WSB_DIR
from webscrapbook import util
from webscrapbook.scrapbook.host import Host
from webscrapbook.scrapbook.convert import file2wsb

root_dir = os.path.abspath(os.path.dirname(__file__))
test_root = os.path.join(root_dir, 'test_scrapbook_convert')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.scrapbook.host.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_CONFIG', test_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestRun(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192
        cls.test_input = os.path.join(test_root, 'input')
        cls.test_output = os.path.join(test_root, 'output')
        cls.test_output_tree = os.path.join(cls.test_output, 'tree')
        cls.test_output_meta = os.path.join(cls.test_output_tree, 'meta.js')
        cls.test_output_toc = os.path.join(cls.test_output_tree, 'toc.js')

    def setUp(self):
        """Set up a general temp test folder
        """
        os.makedirs(self.test_input, exist_ok=True)
        os.makedirs(self.test_output, exist_ok=True)

    def tearDown(self):
        """Remove general temp test folder
        """
        try:
            shutil.rmtree(self.test_input)
        except NotADirectoryError:
            os.remove(self.test_input)
        except FileNotFoundError:
            pass

        try:
            shutil.rmtree(self.test_output)
        except NotADirectoryError:
            os.remove(self.test_output)
        except FileNotFoundError:
            pass

    def test_path01(self):
        """Test hierarchical folders for */index.html
        """
        index_file = os.path.join(self.test_input, 'folder1#中文', 'folder2', 'folder_data', 'index.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<!DOCTYPE html>
<html
    data-scrapbook-create="20200101000000000"
    data-scrapbook-modify="20200101000000000"
    data-scrapbook-source="http://example.com">
<head>
<meta charset="UTF-8">
<title>MyTitle 中文</title>
</head>
<body>
page content
</body>
</html>
""")

        for info in file2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        ids = list(book.meta.keys())
        id_folder1 = ids[0]
        id_folder2 = ids[1]
        id_item = ids[2]
        self.assertDictEqual(book.meta, {
            id_folder1: {
                'title': 'folder1#中文',
                'type': 'folder',
                'create': id_folder1,
                'modify': id_folder1,
                },
            id_folder2: {
                'title': 'folder2',
                'type': 'folder',
                'create': id_folder2,
                'modify': id_folder2,
                },
            id_item: {
                'title': 'MyTitle 中文',
                'type': '',
                'index': f'{id_item}/index.html',
                'create': '20200101000000000',
                'modify': '20200101000000000',
                'source': 'http://example.com',
                'icon': '',
                'comment': '',
                },
            })
        self.assertDictEqual(book.toc, {
            'root': [
                id_folder1,
                ],
            id_folder1: [
                id_folder2,
                ],
            id_folder2: [
                id_item,
                ],
            })
        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, id_item),
            os.path.join(self.test_output, id_item, 'index.html'),
            })

    def test_path02(self):
        """Test hierarchical folders for *.html
        """
        index_file = os.path.join(self.test_input, 'folder1#中文', 'folder2', 'mypage.html')
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<!DOCTYPE html>
<html
    data-scrapbook-create="20200101000000000"
    data-scrapbook-modify="20200101000000000"
    data-scrapbook-source="http://example.com">
<head>
<meta charset="UTF-8">
<title>MyTitle 中文</title>
</head>
<body>
page content
</body>
</html>
""")

        for info in file2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        ids = list(book.meta.keys())
        id_folder1 = ids[0]
        id_folder2 = ids[1]
        id_item = ids[2]
        self.assertDictEqual(book.meta, {
            id_folder1: {
                'title': 'folder1#中文',
                'type': 'folder',
                'create': id_folder1,
                'modify': id_folder1,
                },
            id_folder2: {
                'title': 'folder2',
                'type': 'folder',
                'create': id_folder2,
                'modify': id_folder2,
                },
            id_item: {
                'title': 'MyTitle 中文',
                'type': '',
                'index': f'{id_item}.html',
                'create': '20200101000000000',
                'modify': '20200101000000000',
                'source': 'http://example.com',
                'icon': '',
                'comment': '',
                },
            })
        self.assertDictEqual(book.toc, {
            'root': [
                id_folder1,
                ],
            id_folder1: [
                id_folder2,
                ],
            id_folder2: [
                id_item,
                ],
            })
        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, f'{id_item}.html'),
            })

    def test_path03(self):
        """<input dir>/index.html should be indexed as single html page
        """
        index_file = os.path.join(self.test_input, 'index.html')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<!DOCTYPE html>
<html
    data-scrapbook-create="20200101000000000"
    data-scrapbook-modify="20200101000000000"
    data-scrapbook-source="http://example.com">
<head>
<meta charset="UTF-8">
<title>MyTitle 中文</title>
</head>
<body>
page content
</body>
</html>
""")

        for info in file2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        ids = list(book.meta.keys())
        id_item = ids[0]
        self.assertDictEqual(book.meta, {
            id_item: {
                'title': 'MyTitle 中文',
                'type': '',
                'index': f'{id_item}.html',
                'create': '20200101000000000',
                'modify': '20200101000000000',
                'source': 'http://example.com',
                'icon': '',
                'comment': '',
                },
            })
        self.assertDictEqual(book.toc, {
            'root': [
                id_item,
                ],
            })
        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, f'{id_item}.html'),
            })

    def test_supporting_folder01(self):
        """Test for supporting folder *.files
        """
        index_file = os.path.join(self.test_input, 'mypage.html')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<!DOCTYPE html>
<html
    data-scrapbook-create="20200101000000000"
    data-scrapbook-modify="20200101000000000"
    data-scrapbook-source="http://example.com">
<head>
<meta charset="UTF-8">
<title>MyTitle 中文</title>
</head>
<body>
page content
<img src="mypage.files/picture.bmp">
</body>
</html>
""")
        img_file = os.path.join(self.test_input, 'mypage.files', 'picture.bmp')
        os.makedirs(os.path.dirname(img_file), exist_ok=True)
        with open(img_file, 'wb') as fh:
            fh.write(b'dummy')

        for info in file2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        ids = list(book.meta.keys())
        id_item = ids[0]
        self.assertDictEqual(book.meta, {
            id_item: {
                'title': 'MyTitle 中文',
                'type': '',
                'index': f'{id_item}/index.html',
                'create': '20200101000000000',
                'modify': '20200101000000000',
                'source': 'http://example.com',
                'icon': '',
                'comment': '',
                },
            })
        self.assertDictEqual(book.toc, {
            'root': [
                id_item,
                ],
            })
        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, id_item),
            os.path.join(self.test_output, id_item, 'index.html'),
            os.path.join(self.test_output, id_item, 'mypage.html'),
            os.path.join(self.test_output, id_item, 'mypage.files'),
            os.path.join(self.test_output, id_item, 'mypage.files', 'picture.bmp'),
            })

    def test_supporting_folder02(self):
        """Test for supporting folder *_files
        """
        index_file = os.path.join(self.test_input, 'mypage.html')
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<!DOCTYPE html>
<html
    data-scrapbook-create="20200101000000000"
    data-scrapbook-modify="20200101000000000"
    data-scrapbook-source="http://example.com">
<head>
<meta charset="UTF-8">
<title>MyTitle 中文</title>
</head>
<body>
page content
<img src="mypage_files/picture.bmp">
</body>
</html>
""")
        img_file = os.path.join(self.test_input, 'mypage_files', 'picture.bmp')
        os.makedirs(os.path.dirname(img_file), exist_ok=True)
        with open(img_file, 'wb') as fh:
            fh.write(b'dummy')

        for info in file2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        ids = list(book.meta.keys())
        id_item = ids[0]
        self.assertDictEqual(book.meta, {
            id_item: {
                'title': 'MyTitle 中文',
                'type': '',
                'index': f'{id_item}/index.html',
                'create': '20200101000000000',
                'modify': '20200101000000000',
                'source': 'http://example.com',
                'icon': '',
                'comment': '',
                },
            })
        self.assertDictEqual(book.toc, {
            'root': [
                id_item,
                ],
            })
        self.assertEqual(set(glob.iglob(os.path.join(self.test_output, '**'), recursive=True)), {
            os.path.join(self.test_output, ''),
            os.path.join(self.test_output, id_item),
            os.path.join(self.test_output, id_item, 'index.html'),
            os.path.join(self.test_output, id_item, 'mypage.html'),
            os.path.join(self.test_output, id_item, 'mypage_files'),
            os.path.join(self.test_output, id_item, 'mypage_files', 'picture.bmp'),
            })

if __name__ == '__main__':
    unittest.main()



"""Command line interface of WebScrapBook toolkit.
"""
import sys
import os
import shutil
import argparse
from getpass import getpass
import traceback
from inspect import getdoc

# this package
from . import __package_name__, __version__
from . import *
from . import server
from . import util
from ._compat.time import time_ns


def log(*args):
    print(*args)


def die(*args):
    print('Error:', *args, file=sys.stderr)
    sys.exit(1)


def get_umask():
    """Get configured umask.
    """
    umask = os.umask(0)
    os.umask(umask)
    return umask


def fcopy(fsrc, fdst):
    """Copy a script file to target

    - Auto generate ancestor directories.
    - Use universal linefeed.
    - Set last modified time to current time.
    """
    os.makedirs(os.path.dirname(os.path.abspath(fdst)), exist_ok=True)
    with open(fsrc, 'r', encoding='UTF-8') as fr:
        with open(fdst, 'w', encoding='UTF-8') as fw:
            for line in fr:
                fw.write(line)


def cmd_serve(args):
    """Serve the root directory forever. Shutdown via Ctrl+C or another killing
    technique.

    By default the local browser will be launched to view the served (hosted) site.
    This behavior can be changed using the `server.browse` config.

    Note that this built-in server is designed only for local hosting, or remote
    hosting for personal or few people usage. For an opened world wide web hosting,
    a more specialized server should be used."""
    server.serve(**args)


def cmd_config(args):
    """Show, generate, or edit the config.

    Display the current config when used with no arguments.

    Run `wsb help config` for details about config.
    """
    if args['book']:
        fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, WSB_CONFIG))
        fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'config.ini'))
        if not os.path.isfile(fdst):
            log(f'Generating "{fdst}"...')
            try:
                fcopy(fsrc, fdst)
            except OSError:
                die(f"Unable to generate {fdst}.")

        if args['edit']:
            try:
                util.launch(fdst)
            except OSError:
                pass

        if args['all']:
            fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, 'serve.py'))
            fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'serve.py'))
            if not os.path.isfile(fdst):
                log(f'Generating "{fdst}"...')
                try:
                    fcopy(fsrc, fdst)
                    os.chmod(fdst, os.stat(fdst).st_mode | (0o111 & ~get_umask()))
                except OSError:
                    die(f"Unable to generate {fdst}.")

            fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, 'app.py'))
            fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'app.py'))
            if not os.path.isfile(fdst):
                log(f'Generating "{fdst}"...')
                try:
                    fcopy(fsrc, fdst)
                    os.chmod(fdst, os.stat(fdst).st_mode | (0o111 & ~get_umask()))
                except OSError:
                    die(f"Unable to generate {fdst}.")

    elif args['user']:
        fdst = WSB_USER_CONFIG
        fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'config.ini'))
        if not os.path.isfile(fdst):
            log(f'Generating "{fdst}"...')
            try:
                fcopy(fsrc, fdst)
            except OSError:
                die(f"Unable to generate {fdst}.")

        if args['edit']:
            try:
                util.launch(fdst)
            except OSError:
                pass

    elif args['edit']:
        die("Use --edit in combine with --book or --user.")

    elif args['all']:
        die("Use --all in combine with --book.")

    elif args['name']:
        config.load(args['root'])
        value = config.getname(args['name'])

        if value is None:
            die(f"""Config entry "{args['name']}" does not exist""")

        print(value)

    else:
        config.load(args['root'])
        config.dump(sys.stdout)


def cmd_encrypt(args):
    """Generate an encrypted password string.

    Primilarly to be used in auth config.
    """
    if args['password'] is None:
        pw1 = getpass('Enter a password: ')
        pw2 = getpass('Confirm the password: ')

        if pw1 != pw2:
            die('Entered passwords do not match.')

        args['password'] = pw1

    print(util.encrypt(args['password'], salt=args['salt'], method=args['method']))


def cmd_cache(args):
    """Generate (or update) fulltext cache and/or static site pages.
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import cache
    for info in cache.generate(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_export(args):
    """Export data items into archive files (*.wsba).
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import exporter
    for info in exporter.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_import(args):
    """Import data items from archive files (*.wsba).

    To faithfully recover original tree structure, import items using the same
    order as how they have been exported.
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import importer
    for info in importer.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_check(args):
    """Integrity check and fix for scrapbook data.

    (TOC = table of contents)
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import check
    for info in check.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_convert(args):
    """Convert data between different formats.

    Do not perform any data operation for the input and/or output directory
    during the conversion process (and related backend server(s) should be shut
    down in prior) to prevent a potential conversion error.
    """
    kwargs = args.copy()
    kwargs.pop('root')
    mode = kwargs.pop('mode')
    force = kwargs.pop('force')
    debug = kwargs.pop('debug')

    import importlib
    try:
        conv = importlib.import_module(f'.scrapbook.convert.{mode}', __package__)
    except ImportError:
        die(f'Unsupported conversion mode: "{mode}".')

    # validate input and output directory
    args['input'] = os.path.realpath(args['input'])
    args['output'] = os.path.realpath(args['output'])

    if os.path.normcase(args['output']) == os.path.normcase(args['input']):
        die(f'''Unable to output to the input directory''')

    if os.path.normcase(args['output']).startswith(os.path.normcase(os.path.join(args['input'], ''))):
        die(f'''Unable to output to a descendant of the input directory''')

    if os.path.normcase(args['input']).startswith(os.path.normcase(os.path.join(args['output'], ''))):
        die(f'''Unable to output to an ancestor of the input directory''')

    if not os.path.isdir(args['input']):
        die(f'''Input directory not available: "{args['input']}"''')

    if not os.path.lexists(args['output']):
        pass
    elif not os.path.isdir(args['output']):
        die(f'''Output directory not available: "{args['output']}"''')
    else:
        if force:
            # using os.rmtree() frequently cause an error on Windows
            with os.scandir(args['output']) as dirs:
                for entry in dirs:
                    try:
                        shutil.rmtree(entry)
                    except NotADirectoryError:
                        os.remove(entry)
        else:
            with os.scandir(args['output']) as dirs:
                if next(dirs, None):
                    die(f'''Output directory not empty: "{args['output']}"''')

    for info in conv.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_help(args):
    """Show detailed information about certain topics.
    """
    root = os.path.join(os.path.dirname(__file__), 'resources')

    if args['topic'] == 'config':
        file = os.path.join(root, 'config.md')
        with open(file, 'r', encoding='UTF-8') as f:
            text = f.read()
        print(text)


def cmd_view(args):
    """View archive file(s) in the browser.
    """
    config.load(args['root'])
    view_archive_files(args['files'])


def view_archive_files(files):
    """View archive file(s) in the browser.

    Set default application of MAFF/HTZ archive files to this command to open
    them in the browser directly.
    """
    import tempfile
    import zipfile
    import mimetypes
    import webbrowser
    from urllib.request import pathname2url

    cache_prefix = config['browser']['cache_prefix']
    cache_expire = config['browser']['cache_expire'] * 10 ** 9
    use_jar = config['browser']['use_jar']
    browser = webbrowser.get(config['browser']['command'] or None)

    temp_dir = tempfile.gettempdir()
    urls = []

    for file in dict.fromkeys(os.path.normcase(os.path.abspath(file)) for file in files):
        mime, _ = mimetypes.guess_type(file)
        if mime not in ("application/html+zip", "application/x-maff"):
            continue

        if use_jar:
            base_url = 'jar:file:' + pathname2url(file) + '!/'
            if mime == "application/html+zip":
                urls.append(base_url + 'index.html')
            elif mime == "application/x-maff":
                urls.extend(base_url + f.indexfilename for f in util.get_maff_pages(file))
            continue

        # extract zip contents to dest_dir if not done yet
        hash = util.checksum(file)
        dest_prefix = cache_prefix + hash + '_'
        with os.scandir(temp_dir) as entries:
            for entry in entries:
                if not entry.name.startswith(dest_prefix):
                    continue

                dest_dir = entry.path

                # update atime
                atime = time_ns()
                stat = os.stat(entry)
                os.utime(entry, ns=(atime, stat.st_mtime_ns))
                break
            else:
                dest_dir = tempfile.mkdtemp(prefix=dest_prefix)
                with zipfile.ZipFile(file) as zip:
                    zip.extractall(dest_dir)

        # get URL of every index page
        base_url = 'file:' + pathname2url(dest_dir) + '/'
        if mime == "application/html+zip":
            urls.append(base_url + 'index.html')
        elif mime == "application/x-maff":
            urls.extend(base_url + f.indexfilename for f in util.get_maff_pages(file))

    # open pages in the browser
    for url in urls:
        browser.open(url)

    # remove stale caches
    if not use_jar:
        t = time_ns()
        with os.scandir(temp_dir) as entries:
            for entry in entries:
                if not entry.name.startswith(cache_prefix):
                    continue

                atime = os.stat(entry).st_atime_ns
                if t <= atime + cache_expire:
                    continue

                # cache may be created by another user and undeletable
                try:
                    shutil.rmtree(entry)
                except OSError:
                    traceback.print_exc()


def view():
    """CLI entry point for viewing archive files.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(view_archive_files))
    parser.add_argument('files', nargs='+',
        help="""files to view.""")
    args = vars(parser.parse_args())
    view_archive_files(args['files'])


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--version', action='version', version=f'{__package_name__} {__version__}',
        help="""show version information and exit""")
    parser.add_argument('--root', default=".",
        help="""root directory to manipulate (default: current working directory)""")
    subparsers = parser.add_subparsers(metavar='COMMAND',
        help="""the sub-command to run. Get usage help with e.g. %(prog)s config -h""")

    # subcommand: serve
    parser_serve = subparsers.add_parser('serve', aliases=['s'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_serve),
        help="""serve the root directory""")
    parser_serve.add_argument('--browse', default=None, action='store_true',
        help="""launch the browser to visit the served directory""")
    parser_serve.add_argument('--no-browse', dest='browse', action='store_false',
        help="""do not launch the browser""")
    parser_serve.set_defaults(func=cmd_serve)

    # subcommand: config
    parser_config = subparsers.add_parser('config', aliases=['c'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_config),
        help="""show, generate, or edit the config""")
    parser_config.set_defaults(func=cmd_config)
    parser_config.add_argument('name', nargs='?',
        help="""show value of the given config name (in the form of <section>[.<subsection>].<key>)""")
    parser_config.add_argument('-b', '--book', default=False, action='store_true',
        help="""generate book (host) config file""")
    parser_config.add_argument('-u', '--user', default=False, action='store_true',
        help="""generate user config file""")
    parser_config.add_argument('-a', '--all', default=False, action='store_true',
        help="""generate more assistant files (with --book)""")
    parser_config.add_argument('-e', '--edit', default=False, action='store_true',
        help="""edit the config file (with --book or --user)""")

    # subcommand: encrypt
    parser_encrypt = subparsers.add_parser('encrypt', aliases=['e'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_encrypt),
        help="""generate an encrypted password""")
    parser_encrypt.set_defaults(func=cmd_encrypt)
    parser_encrypt.add_argument('-p', '--password', nargs='?', default=None, action='store',
        help="""the password to encrypt. Skip to provide via an interactive prompt.""")
    parser_encrypt.add_argument('-m', '--method', default='sha1', action='store',
        help="""the encrypt method to use, which is one of: plain, md5, sha1,
sha224, sha256, sha384, sha512, sha3_224, sha3_256, sha3_384, and sha3_512
(default: %(default)s)""")
    parser_encrypt.add_argument('-s', '--salt', default='', action='store',
        help="""the salt to add during encryption.""")

    # subcommand: cache
    parser_cache = subparsers.add_parser('cache', aliases=['a'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_cache),
        help="""update fulltext cache and/or static site pages""")
    parser_cache.set_defaults(func=cmd_cache)
    parser_cache.add_argument('book_ids', metavar='book', nargs='*', action='store',
        help="""the book ID(s) to generate cache. (default: all books)""")
    parser_cache.add_argument('--item', dest='item_ids',
        metavar='ID', action='store', default=None, nargs='+',
        help="""the items ID(s) to generate cache (default: all)""")
    parser_cache.add_argument('--fulltext', default=True, action='store_true',
        help="""generate fulltext cache. (default)""")
    parser_cache.add_argument('--no-fulltext', dest='fulltext', action='store_false',
        help="""inverse of --fulltext""")
    parser_cache.add_argument('--inclusive-frames', default=True, action='store_true',
        help="""cache frame content as part of the main page (default). It's
recommended to recreate fulltext cache when changing this option to prevent
inconsistency.""")
    parser_cache.add_argument('--no-inclusive-frames', dest='inclusive_frames', action='store_false',
        help="""inverse of --inclusive-frames""")
    parser_cache.add_argument('--recreate', dest='recreate', default=False, action='store_true',
        help="""ignore current fulltext cache and generate again""")
    parser_cache.add_argument('--no-recreate', dest='recreate', action='store_false',
        help="""inverse of --recreate (default)""")
    parser_cache.add_argument('--static-site', default=False, action='store_true',
        help="""generate static site pages""")
    parser_cache.add_argument('--no-static-site', dest='static_site', action='store_false',
        help="""inverse of --static-site (default)""")
    parser_cache.add_argument('--static-index', default=False, action='store_true',
        help="""generate static index.html page""")
    parser_cache.add_argument('--no-static-index', dest='static_index', action='store_false',
        help="""inverse of --static-index (default)""")
    parser_cache.add_argument('--rss-root', metavar='ROOT_URL', action='store',
        help="""generate an RSS feed file for the book, using the specified root URL
        (usually corresponds to webscrapbook app root)""")
    parser_cache.add_argument('--rss-item-count', default=50, type=int, action='store',
        help="""number of items the RSS feed should include (default: %(default)s)""")
    parser_cache.add_argument('--locale', action='store',
        help="""locale for the generated pages (default: system locale)""")
    parser_cache.add_argument('--backup', dest='no_backup', default=True, action='store_false',
        help="""backup changed files""")
    parser_cache.add_argument('--no-backup', action='store_true',
        help="""do not backup changed files (default)""")
    parser_cache.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: check
    parser_check = subparsers.add_parser('check', aliases=['k'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_check),
        help="""check and fix scrapbook data""")
    parser_check.set_defaults(func=cmd_check)
    parser_check.add_argument('book_ids', metavar='book', nargs='*', action='store',
        help="""the book ID(s) to check. (default: all books)""")

    parser_check.add_argument('-r', '--resolve', dest='resolve_all', default=False, action='store_true',
        help="""resolve all found issues (implies all --resolve-*)""")
    parser_check.add_argument('--resolve-invalid-id', default=False, action='store_true',
        help="""remove items with invalid ID from metadata entries""")
    parser_check.add_argument('--resolve-missing-index', default=False, action='store_true',
        help="""remove items with missing index property from metadata entries""")
    parser_check.add_argument('--resolve-missing-index-file', default=False, action='store_true',
        help="""remove items with missing index file from metadata entries""")
    parser_check.add_argument('--resolve-missing-date', default=False, action='store_true',
        help="""attempt to generate "create" and "modify" properties for items missing any of them""")
    parser_check.add_argument('--resolve-older-mtime', default=False, action='store_true',
        help="""update "modify" property if it's older than last modified time of the index file""")
    parser_check.add_argument('--resolve-toc-unreachable', default=False, action='store_true',
        help="""append items unreachable from TOC to the root tree""")
    parser_check.add_argument('--resolve-toc-invalid', default=False, action='store_true',
        help="""remove invalid items from TOC""")
    parser_check.add_argument('--resolve-toc-empty-subtree', default=False, action='store_true',
        help="""remove items with empty subtree from TOC""")
    parser_check.add_argument('--resolve-unindexed-files', default=False, action='store_true',
        help="""attempt to import unindexed files to metadata and TOC""")
    parser_check.add_argument('--resolve-absolute-icon', default=False, action='store_true',
        help="""cache "icon" property with absolute URL to local favicon directory""")
    parser_check.add_argument('--resolve-unused-icon', default=False, action='store_true',
        help="""remove unused favicon caches""")

    parser_check.add_argument('--no-backup', default=False, action='store_true',
        help="""do not backup changed files""")
    parser_check.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: export
    parser_export = subparsers.add_parser('export', aliases=['x'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_export),
        help="""export data items into archive files (*.wsba)""")
    parser_export.set_defaults(func=cmd_export)
    parser_export.add_argument('output', action='store',
        help="""the output directory""")
    parser_export.add_argument('--book', dest='book_id', metavar='ID', default='', action='store',
        help="""the book ID to export. (default: "")""")
    parser_export.add_argument('--item', dest='item_ids',
        metavar='ID', action='store', default=None, nargs='+',
        help="""the items ID(s) to export (default: all)""")
    parser_export.add_argument('-r', '--recursive', default=False, action='store_true',
        help="""recursively include descendant items of the provided item ID(s)""")
    parser_export.add_argument('-s', '--singleton', default=False, action='store_true',
        help="""export only the first occurrence for an item referenced many times""")
    parser_export.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: import
    parser_import = subparsers.add_parser('import', aliases=['i'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_import),
        help="""import data items from archive files (*.wsba)""",
        epilog="""\
Available MODEs for --resolve-id-used:
  "skip": skip the import
  "replace": replace current metadata and data file(s)
  "new": import as a new item

Available placeholders for --filename PATTERN:
  "%%": a literal "%"
  "%ID%": item ID
  "%EID%": item export ID
  "%UUID%": a random UUID
  "%TITLE%": item title
  "%SOURCE%": item source URL
  "%CREATE%": item create time (allow SUBPATTERN)
  "%MODIFY%": item modify time (allow SUBPATTERN)
  "%EXPORT%": item export time (allow SUBPATTERN)

Datetime related ones can be written as %PATTERN:SUBPATTERN% for further
formatting with SUBPATTERN being one of these placeholders:
  "DATE": date (YYYY-MM-DD)
  "TIME": time (HH-MM-SS)
  "YEAR": year (YYYY)
  "MONTH": year (MM, 01-12)
  "DAY": day of month (DD, 01-31)
  "HOURS": hours (HH, 01-24)
  "MINUTES": minutes (MM, 00-59)
  "SECONDS": seconds (SS, 00-59)

All SUBPATTERNs can be prepended with "UTC_" for UTC time instead of local
time. For example, "%CREATE:UTC_DATE%".
""")
    parser_import.set_defaults(func=cmd_import)
    parser_import.add_argument('files', metavar='file', action='store', nargs='+',
        help="""the file(s) to import in order. If a directory is provided, all
child files are imported in unicode filename order.""")
    parser_import.add_argument('--book', dest='book_id', metavar='ID', default='', action='store',
        help="""the book ID to import into. (default: "")""")
    parser_import.add_argument('--target', dest='target_id', metavar='ID',
        default='root', action='store',
        help="""the target item ID to insert the imported items under (default: "%(default)s")""")
    parser_import.add_argument('--target-index', metavar='INDEX',
        type=int, action='store',
        help="""the index number (starting from 0) the imported items will be
inserted at (default: last)""")
    parser_import.add_argument('--rebuild-folders', default=False, action='store_true',
        help="""insert imported items under the original parent, and
auto-generate parent folders if not found. (ignores --target and
--target-index)""")
    parser_import.add_argument('--resolve-id-used', metavar='MODE',
        default='skip', action='store',
        choices={'skip', 'replace', 'new'},
        help="""what to do if an importing item ID already exists (default: "%(default)s")""")
    parser_import.add_argument('--filename', dest='target_filename', metavar='PATTERN',
        default='%ID%', action='store',
        help="""formatter of the imported filename (default: "%(default)s")""")
    parser_import.add_argument('--prune', default=False, action='store_true',
        help="""delete the archive file after successfully imported""")
    parser_import.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: convert
    parser_convert = subparsers.add_parser('convert', aliases=['v'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_convert),
        help="""convert scrapbook data between different formats""")
    parser_convert.set_defaults(func=cmd_convert)
    parser_convert_sub = parser_convert.add_subparsers(dest='mode', metavar='MODE',
        help="""the conversion mode. Get usage help with e.g. %(prog)s sb2wsb -h""")

    # -- sb2wsb
    parser_convert_sb2wsb = parser_convert_sub.add_parser('sb2wsb',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Convert from legacy ScrapBook (X) to WebScrapBook.""",
        help="""convert from legacy ScrapBook (X) to WebScrapBook""")
    parser_convert_sb2wsb.add_argument('input', action='store',
        help="""the input directory""")
    parser_convert_sb2wsb.add_argument('output', action='store',
        help="""the output directory""")
    parser_convert_sb2wsb.add_argument('--no-backup', default=False, action='store_true',
        help="""do not backup unneeded legacy scrapbook files""")
    parser_convert_sb2wsb.add_argument('--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_sb2wsb.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- wsb2sb
    parser_convert_wsb2sb = parser_convert_sub.add_parser('wsb2sb',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Convert from WebScrapBook to legacy ScrapBook X

Note that certain information may lose permanently, such as:
* item appended to multiple parents (preserve only the first occurence)
* item in the recycle bin

Also note that compatibility validation of this tool is targeting ScrapBook X.
There may be minor compatibility issues if the output scrapbook is used by a
legacy ScrapBook implementation without features introduced by ScrapBook X,
such as:
* file with special or non-ASCII chars in filename
* container item whose type property is not "folder" """,
        help="""convert from WebScrapBook to legacy ScrapBook X""")
    parser_convert_wsb2sb.add_argument('input', action='store',
        help="""the input directory""")
    parser_convert_wsb2sb.add_argument('output', action='store',
        help="""the output directory""")
    parser_convert_wsb2sb.add_argument('--book', dest='book_id', metavar='ID',
        default='', action='store',
        help="""the book ID to convert. (default: "")""")
    parser_convert_wsb2sb.add_argument('--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_wsb2sb.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- file2wsb
    parser_convert_file2wsb = parser_convert_sub.add_parser('file2wsb',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Convert from hierarchical files to WebScrapBook.

Supports metadata recognition of web pages saved via:
- Built-in save of Internet Explorer or a Chromium-based browser
- SingleFile
- Save Page WE
- MaoXian web clipper
""",
        help="""convert from hierarchical files to WebScrapBook""")
    parser_convert_file2wsb.add_argument('input', action='store',
        help="""the input directory""")
    parser_convert_file2wsb.add_argument('output', action='store',
        help="""the output directory""")
    parser_convert_file2wsb.add_argument('--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_file2wsb.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- wsb2file
    parser_convert_wsb2file = parser_convert_sub.add_parser('wsb2file',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Convert from WebScrapBook to hierarchical files""",
        help="""convert from WebScrapBook to hierarchical files""")
    parser_convert_wsb2file.add_argument('input', action='store',
        help="""the input directory""")
    parser_convert_wsb2file.add_argument('output', action='store',
        help="""the output directory""")
    parser_convert_wsb2file.add_argument('--book', dest='book_id', metavar='ID',
        default='', action='store',
        help="""the book ID to convert. (default: "")""")
    parser_convert_wsb2file.add_argument('--no-prefix', dest='prefix', default=True, action='store_false',
        help="""don't prefix output files with position number.""")
    parser_convert_wsb2file.add_argument('--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_wsb2file.add_argument('--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: help
    parser_help = subparsers.add_parser('help',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_help),
        help="""show detailed information about certain topics""")
    parser_help.set_defaults(func=cmd_help)
    parser_help.add_argument('topic', default=None, action='store',
        choices=['config'],
        help="""the topic for details""")

    # subcommand: view
    parser_view = subparsers.add_parser('view',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_view),
        help="""view archive file in the browser""")
    parser_view.set_defaults(func=cmd_view)
    parser_view.add_argument('files', nargs='+',
        help="""files to view""")

    # parse the command
    args = vars(parser.parse_args())
    try:
        func = args.pop('func')
    except KeyError:
        parser.parse_args(['-h'])
    else:
        if func is cmd_convert and args['mode'] is None:
            parser.parse_args(['convert', '-h'])
            return

        func(args)


if __name__ == '__main__':
    main()



from unittest import mock
import unittest
import os
import shutil
import io
import re
import zipfile
import time
import functools
from webscrapbook import WSB_DIR, Config
from webscrapbook import util
from webscrapbook.scrapbook.host import Host
from webscrapbook.scrapbook import book as wsb_book
from webscrapbook.scrapbook.book import Book

root_dir = os.path.abspath(os.path.dirname(__file__))
test_root = os.path.join(root_dir, 'test_scrapbook_book')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.scrapbook.host.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_CONFIG', test_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestBook(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192
        cls.test_root = os.path.join(test_root, 'general')
        cls.test_wsbdir = os.path.join(cls.test_root, WSB_DIR)
        cls.test_config = os.path.join(cls.test_root, WSB_DIR, 'config.ini')

    def setUp(self):
        """Set up a general temp test folder
        """
        try:
            shutil.rmtree(self.test_root)
        except NotADirectoryError:
            os.remove(self.test_root)
        except FileNotFoundError:
            pass

        os.makedirs(self.test_wsbdir)

    def tearDown(self):
        """Remove general temp test folder
        """
        try:
            shutil.rmtree(self.test_root)
        except NotADirectoryError:
            os.remove(self.test_root)
        except FileNotFoundError:
            pass

    def create_general_config(self):
        with open(self.test_config, 'w', encoding='UTF-8') as f:
            f.write("""[book ""]
name = scrapbook
top_dir =
data_dir = data
tree_dir = tree
index = tree/map.html
no_tree = false
""")

    def test_init01(self):
        """Check basic"""
        with open(self.test_config, 'w', encoding='UTF-8') as f:
            f.write("""[book ""]
name = scrapbook
top_dir = 
data_dir = data
tree_dir = tree
index = tree/map.html
no_tree = false
""")

        host = Host(self.test_root)
        book = Book(host)

        self.assertEqual(book.host, host)
        self.assertEqual(book.id, '')
        self.assertEqual(book.name, 'scrapbook')
        self.assertEqual(book.root, self.test_root)
        self.assertEqual(book.top_dir, self.test_root)
        self.assertEqual(book.data_dir, os.path.join(self.test_root, 'data'))
        self.assertEqual(book.tree_dir, os.path.join(self.test_root, 'tree'))
        self.assertFalse(book.no_tree)

    def test_init02(self):
        """Check book_id param"""
        with open(self.test_config, 'w', encoding='UTF-8') as f:
            f.write("""[book "book1"]
name = scrapbook1
top_dir =
data_dir =
tree_dir = .wsb/tree
index = .wsb/tree/map.html
no_tree = false
""")

        host = Host(self.test_root)
        book = Book(host, 'book1')

        self.assertEqual(book.host, host)
        self.assertEqual(book.id, 'book1')
        self.assertEqual(book.name, 'scrapbook1')
        self.assertEqual(book.root, self.test_root)
        self.assertEqual(book.top_dir, self.test_root)
        self.assertEqual(book.data_dir, self.test_root)
        self.assertEqual(book.tree_dir, os.path.join(self.test_root, '.wsb', 'tree'))
        self.assertFalse(book.no_tree)

    def test_init03(self):
        """Check modified path"""
        with open(self.test_config, 'w', encoding='UTF-8') as f:
            f.write("""[app]
root = public

[book ""]
name = scrapbook
top_dir = sb
data_dir = data
tree_dir = tree
index = tree/map.html
no_tree = false
""")

        host = Host(self.test_root)
        book = Book(host)

        self.assertEqual(book.host, host)
        self.assertEqual(book.id, '')
        self.assertEqual(book.name, 'scrapbook')
        self.assertEqual(book.root, self.test_root)
        self.assertEqual(book.top_dir, os.path.join(self.test_root, 'public', 'sb'))
        self.assertEqual(book.data_dir, os.path.join(self.test_root, 'public', 'sb', 'data'))
        self.assertEqual(book.tree_dir, os.path.join(self.test_root, 'public', 'sb', 'tree'))
        self.assertFalse(book.no_tree)

    def test_get_subpath(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        self.assertEqual(book.get_subpath(os.path.join(self.test_root, 'tree', 'meta.js')), 'tree/meta.js')

    def test_get_tree_file(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        self.assertEqual(book.get_tree_file('meta'), os.path.join(self.test_root, 'tree', 'meta.js'))
        self.assertEqual(book.get_tree_file('toc', 1), os.path.join(self.test_root, 'tree', 'toc1.js'))

    def test_iter_tree_files01(self):
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'tree'))
        with open(os.path.join(self.test_root, 'tree', 'meta.js'), 'w', encoding='UTF-8') as f:
            pass
        with open(os.path.join(self.test_root, 'tree', 'meta1.js'), 'w', encoding='UTF-8') as f:
            pass
        with open(os.path.join(self.test_root, 'tree', 'meta2.js'), 'w', encoding='UTF-8') as f:
            pass

        book = Book(Host(self.test_root))
        self.assertEqual(list(book.iter_tree_files('meta')), [
            os.path.join(self.test_root, 'tree', 'meta.js'),
            os.path.join(self.test_root, 'tree', 'meta1.js'),
            os.path.join(self.test_root, 'tree', 'meta2.js'),
            ])

    def test_iter_tree_files02(self):
        """Break since nonexisting index"""
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'tree'))
        with open(os.path.join(self.test_root, 'tree', 'meta.js'), 'w', encoding='UTF-8') as f:
            pass
        with open(os.path.join(self.test_root, 'tree', 'meta1.js'), 'w', encoding='UTF-8') as f:
            pass
        with open(os.path.join(self.test_root, 'tree', 'meta3.js'), 'w', encoding='UTF-8') as f:
            pass

        book = Book(Host(self.test_root))
        self.assertEqual(list(book.iter_tree_files('meta')), [
            os.path.join(self.test_root, 'tree', 'meta.js'),
            os.path.join(self.test_root, 'tree', 'meta1.js'),
            ])

    def test_iter_tree_files03(self):
        """Works when directory not exist"""
        book = Book(Host(self.test_root))
        self.assertEqual(list(book.iter_tree_files('meta')), [])

    @mock.patch('webscrapbook.scrapbook.book.Book.iter_tree_files')
    def test_iter_meta_files(self, mock_func):
        book = Book(Host(self.test_root))
        for i in book.iter_meta_files():
            pass
        mock_func.assert_called_once_with('meta')

    @mock.patch('webscrapbook.scrapbook.book.Book.iter_tree_files')
    def test_iter_toc_files(self, mock_func):
        book = Book(Host(self.test_root))
        for i in book.iter_toc_files():
            pass
        mock_func.assert_called_once_with('toc')

    @mock.patch('webscrapbook.scrapbook.book.Book.iter_tree_files')
    def test_iter_fulltext_files(self, mock_func):
        book = Book(Host(self.test_root))
        for i in book.iter_fulltext_files():
            pass
        mock_func.assert_called_once_with('fulltext')

    def test_load_tree_file01(self):
        """Test normal loading"""
        self.create_general_config()
        with open(os.path.join(self.test_root, 'meta.js'), 'w', encoding='UTF-8') as f:
            f.write("""/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.meta({
  "20200101000000000": {
    "index": "20200101000000000/index.html",
    "title": "Dummy",
    "type": "",
    "create": "20200101000000000",
    "modify": "20200101000000000"
  }
})""")

        book = Book(Host(self.test_root))
        self.assertEqual(
            book.load_tree_file(os.path.join(self.test_root, 'meta.js')), {
                '20200101000000000': {
                    'index': '20200101000000000/index.html',
                    'title': 'Dummy',
                    'type': '',
                    'create': '20200101000000000',
                    'modify': '20200101000000000',
                    },
                })

    def test_load_tree_file02(self):
        """Test malformed wrapping"""
        self.create_general_config()
        with open(os.path.join(self.test_root, 'meta.js'), 'w', encoding='UTF-8') as f:
            f.write("""
scrapbook.meta({
  "20200101000000000": {
    "index": "20200101000000000/index.html",
    "title": "Dummy",
    "type": "",
    "create": "20200101000000000",
    "modify": "20200101000000000"
  }
}""")

        book = Book(Host(self.test_root))
        with self.assertRaises(wsb_book.TreeFileMalformedWrappingError):
            book.load_tree_file(os.path.join(self.test_root, 'meta.js'))

    def test_load_tree_file03(self):
        """Test malformed wrapping"""
        self.create_general_config()
        with open(os.path.join(self.test_root, 'meta.js'), 'w', encoding='UTF-8') as f:
            f.write("""
scrapbook.meta{
  "20200101000000000": {
    "index": "20200101000000000/index.html",
    "title": "Dummy",
    "type": "",
    "create": "20200101000000000",
    "modify": "20200101000000000"
  }
})""")

        book = Book(Host(self.test_root))
        with self.assertRaises(wsb_book.TreeFileMalformedWrappingError):
            book.load_tree_file(os.path.join(self.test_root, 'meta.js'))

    def test_load_tree_file04(self):
        """Test malformed wrapping"""
        self.create_general_config()
        with open(os.path.join(self.test_root, 'meta.js'), 'w', encoding='UTF-8') as f:
            f.write("""({
  "20200101000000000": {
    "index": "20200101000000000/index.html",
    "title": "Dummy",
    "type": "",
    "create": "20200101000000000",
    "modify": "20200101000000000"
  }
})""")

        book = Book(Host(self.test_root))
        with self.assertRaises(wsb_book.TreeFileMalformedWrappingError):
            book.load_tree_file(os.path.join(self.test_root, 'meta.js'))

    def test_load_tree_file05(self):
        """Test malformed JSON"""
        self.create_general_config()
        with open(os.path.join(self.test_root, 'meta.js'), 'w', encoding='UTF-8') as f:
            f.write("""
scrapbook.meta({
  '20200101000000000': {
    index: '20200101000000000/index.html',
    title: 'Dummy',
    type: '',
    create: '20200101000000000',
    modify: '20200101000000000'
  }
})""")

        book = Book(Host(self.test_root))
        with self.assertRaises(wsb_book.TreeFileMalformedJsonError):
            book.load_tree_file(os.path.join(self.test_root, 'meta.js'))

    def test_load_tree_files01(self):
        """Test normal loading

        - Item of same ID from the latter overwrites the formatter.
        - Item with None value should be removed.
        """
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'tree'))
        with open(os.path.join(self.test_root, 'tree', 'meta.js'), 'w', encoding='UTF-8') as f:
            f.write("""/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.meta({
  "20200101000000000": {
    "index": "20200101000000000/index.html",
    "title": "Dummy",
    "type": "",
    "create": "20200101000000000",
    "modify": "20200101000000000",
    "comment": "comment"
  },
  "20200101000000001": {
    "index": "20200101000000001/index.html",
    "title": "Dummy1",
    "type": "",
    "create": "20200101000000001",
    "modify": "20200101000000001",
    "comment": "comment1"
  },
  "20200101000000002": {
    "index": "20200101000000002/index.html",
    "title": "Dummy2",
    "type": "",
    "create": "20200101000000002",
    "modify": "20200101000000002",
    "comment": "comment2"
  }
})""")
        with open(os.path.join(self.test_root, 'tree', 'meta1.js'), 'w', encoding='UTF-8') as f:
            f.write("""/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.meta({
  "20200101000000001": {
    "index": "20200101000000001/index.html",
    "title": "Dummy1rev",
    "type": "",
    "create": "20200101000000001",
    "modify": "20200101000000011"
  },
  "20200101000000002": null,
  "20200101000000003": {
    "index": "20200101000000003/index.html",
    "title": "Dummy3",
    "type": "",
    "create": "20200101000000003",
    "modify": "20200101000000003",
    "comment": "comment3"
  }
})""")

        book = Book(Host(self.test_root))
        self.assertEqual(book.load_tree_files('meta'), {
            '20200101000000000': {
                'index': '20200101000000000/index.html',
                'title': 'Dummy',
                'type': '',
                'create': '20200101000000000',
                'modify': '20200101000000000',
                'comment': 'comment',
                },
            '20200101000000001': {
                'index': '20200101000000001/index.html',
                'title': 'Dummy1rev',
                'type': '',
                'create': '20200101000000001',
                'modify': '20200101000000011',
                },
            '20200101000000003': {
                'index': '20200101000000003/index.html',
                'title': 'Dummy3',
                'type': '',
                'create': '20200101000000003',
                'modify': '20200101000000003',
                'comment': 'comment3',
                },
            })

    def test_load_tree_files02(self):
        """Works when directory not exist"""
        book = Book(Host(self.test_root))
        self.assertEqual(book.load_tree_files('meta'), {})

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_meta_files01(self, mock_func):
        book = Book(Host(self.test_root))
        book.load_meta_files()
        mock_func.assert_called_once_with('meta')

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_meta_files02(self, mock_func):
        book = Book(Host(self.test_root))
        book.meta = {}
        book.load_meta_files()
        mock_func.assert_not_called()

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_meta_files03(self, mock_func):
        book = Book(Host(self.test_root))
        book.meta = {}
        book.load_meta_files(refresh=True)
        mock_func.assert_called_once_with('meta')

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_toc_files01(self, mock_func):
        book = Book(Host(self.test_root))
        book.load_toc_files()
        mock_func.assert_called_once_with('toc')

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_toc_files02(self, mock_func):
        book = Book(Host(self.test_root))
        book.toc = {}
        book.load_toc_files()
        mock_func.assert_not_called()

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_toc_files03(self, mock_func):
        book = Book(Host(self.test_root))
        book.toc = {}
        book.load_toc_files(refresh=True)
        mock_func.assert_called_once_with('toc')

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_fulltext_files01(self, mock_func):
        book = Book(Host(self.test_root))
        book.load_fulltext_files()
        mock_func.assert_called_once_with('fulltext')

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_fulltext_files02(self, mock_func):
        book = Book(Host(self.test_root))
        book.fulltext = {}
        book.load_fulltext_files()
        mock_func.assert_not_called()

    @mock.patch('webscrapbook.scrapbook.book.Book.load_tree_files')
    def test_load_fulltext_files03(self, mock_func):
        book = Book(Host(self.test_root))
        book.fulltext = {}
        book.load_fulltext_files(refresh=True)
        mock_func.assert_called_once_with('fulltext')

    def test_save_meta_files01(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        book.meta = {
            '20200101000000000': {'title': 'Dummy 1 中文'},
            '20200101000000001': {'title': 'Dummy 2 中文'},
            }

        book.save_meta_files()

        with open(os.path.join(self.test_root, 'tree', 'meta.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.meta({
  "20200101000000000": {
    "title": "Dummy 1 中文"
  },
  "20200101000000001": {
    "title": "Dummy 2 中文"
  }
})""")

    @mock.patch('webscrapbook.scrapbook.book.Book.SAVE_META_THRESHOLD', 3)
    def test_save_meta_files02(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        book.meta = {
            '20200101000000000': {'title': 'Dummy 1 中文'},
            '20200101000000001': {'title': 'Dummy 2 中文'},
            '20200101000000002': {'title': 'Dummy 3 中文'},
            '20200101000000003': {'title': 'Dummy 4 中文'},
            }

        book.save_meta_files()

        with open(os.path.join(self.test_root, 'tree', 'meta.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.meta({
  "20200101000000000": {
    "title": "Dummy 1 中文"
  },
  "20200101000000001": {
    "title": "Dummy 2 中文"
  }
})""")
        with open(os.path.join(self.test_root, 'tree', 'meta1.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.meta({
  "20200101000000002": {
    "title": "Dummy 3 中文"
  },
  "20200101000000003": {
    "title": "Dummy 4 中文"
  }
})""")

    def test_save_meta_files03(self):
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'tree'))
        with open(os.path.join(self.test_root, 'tree', 'meta.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy')
        with open(os.path.join(self.test_root, 'tree', 'meta1.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy1')
        with open(os.path.join(self.test_root, 'tree', 'meta2.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy2')
        with open(os.path.join(self.test_root, 'tree', 'meta3.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy3')

        book = Book(Host(self.test_root))
        book.meta = {
            '20200101000000000': {'title': 'Dummy 1 中文'},
            '20200101000000001': {'title': 'Dummy 2 中文'},
            }

        book.save_meta_files()

        with open(os.path.join(self.test_root, 'tree', 'meta.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.meta({
  "20200101000000000": {
    "title": "Dummy 1 中文"
  },
  "20200101000000001": {
    "title": "Dummy 2 中文"
  }
})""")
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'meta1.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'meta2.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'meta3.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'meta4.js')))

    def test_save_toc_files01(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        book.toc = {
            'root': [
                '20200101000000000',
                '20200101000000001',
                '20200101000000002',
                ],
            '20200101000000000': [
                '20200101000000003'
                ]
            }

        book.save_toc_files()

        with open(os.path.join(self.test_root, 'tree', 'toc.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.toc({
  "root": [
    "20200101000000000",
    "20200101000000001",
    "20200101000000002"
  ],
  "20200101000000000": [
    "20200101000000003"
  ]
})""")

    @mock.patch('webscrapbook.scrapbook.book.Book.SAVE_TOC_THRESHOLD', 3)
    def test_save_toc_files02(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        book.toc = {
            'root': [
                '20200101000000000',
                '20200101000000001',
                '20200101000000002',
                '20200101000000003',
                '20200101000000004',
                ],
            '20200101000000001': [
                '20200101000000011'
                ],
            '20200101000000002': [
                '20200101000000021'
                ],
            '20200101000000003': [
                '20200101000000031',
                '20200101000000032'
                ],
            }

        book.save_toc_files()

        with open(os.path.join(self.test_root, 'tree', 'toc.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.toc({
  "root": [
    "20200101000000000",
    "20200101000000001",
    "20200101000000002",
    "20200101000000003",
    "20200101000000004"
  ]
})""")
        with open(os.path.join(self.test_root, 'tree', 'toc1.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.toc({
  "20200101000000001": [
    "20200101000000011"
  ],
  "20200101000000002": [
    "20200101000000021"
  ]
})""")
        with open(os.path.join(self.test_root, 'tree', 'toc2.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.toc({
  "20200101000000003": [
    "20200101000000031",
    "20200101000000032"
  ]
})""")

    def test_save_toc_files03(self):
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'tree'))
        with open(os.path.join(self.test_root, 'tree', 'toc.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy')
        with open(os.path.join(self.test_root, 'tree', 'toc1.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy1')
        with open(os.path.join(self.test_root, 'tree', 'toc2.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy2')
        with open(os.path.join(self.test_root, 'tree', 'toc4.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy4')

        book = Book(Host(self.test_root))
        book.toc = {
            'root': [
                '20200101000000000',
                '20200101000000001',
                '20200101000000002',
                ],
            '20200101000000000': [
                '20200101000000003'
                ]
            }

        book.save_toc_files()

        with open(os.path.join(self.test_root, 'tree', 'toc.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.toc({
  "root": [
    "20200101000000000",
    "20200101000000001",
    "20200101000000002"
  ],
  "20200101000000000": [
    "20200101000000003"
  ]
})""")
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'toc1.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'toc2.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'toc3.js')))
        self.assertTrue(os.path.exists(os.path.join(self.test_root, 'tree', 'toc4.js')))

    def test_save_fulltext_files01(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        book.fulltext = {
            "20200101000000000": {
                'index.html': {
                    'content': 'dummy text 1 中文',
                    }
                },
            "20200101000000001": {
                'index.html': {
                    'content': 'dummy text 2 中文',
                    }
                },
            }

        book.save_fulltext_files()

        with open(os.path.join(self.test_root, 'tree', 'fulltext.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.fulltext({
 "20200101000000000": {
  "index.html": {
   "content": "dummy text 1 中文"
  }
 },
 "20200101000000001": {
  "index.html": {
   "content": "dummy text 2 中文"
  }
 }
})""")

    @mock.patch('webscrapbook.scrapbook.book.Book.SAVE_FULLTEXT_THRESHOLD', 10)
    def test_save_fulltext_files02(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        book.fulltext = {
            "20200101000000000": {
                'index.html': {
                    'content': 'dummy text 1 中文',
                    },
                'frame.html': {
                    'content': 'frame page content',
                    },
                },
            "20200101000000001": {
                'index.html': {
                    'content': 'dummy text 2 中文',
                    },
                },
            "20200101000000002": {
                'index.html': {
                    'content': 'dummy text 3 中文',
                    },
                },
            }

        book.save_fulltext_files()

        with open(os.path.join(self.test_root, 'tree', 'fulltext.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.fulltext({
 "20200101000000000": {
  "index.html": {
   "content": "dummy text 1 中文"
  },
  "frame.html": {
   "content": "frame page content"
  }
 }
})""")
        with open(os.path.join(self.test_root, 'tree', 'fulltext1.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.fulltext({
 "20200101000000001": {
  "index.html": {
   "content": "dummy text 2 中文"
  }
 }
})""")
        with open(os.path.join(self.test_root, 'tree', 'fulltext2.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.fulltext({
 "20200101000000002": {
  "index.html": {
   "content": "dummy text 3 中文"
  }
 }
})""")

    @mock.patch('webscrapbook.scrapbook.book.Book.SAVE_FULLTEXT_THRESHOLD', 10)
    def test_save_fulltext_files03(self):
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'tree'))
        with open(os.path.join(self.test_root, 'tree', 'fulltext.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy')
        with open(os.path.join(self.test_root, 'tree', 'fulltext1.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy1')
        with open(os.path.join(self.test_root, 'tree', 'fulltext2.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy2')
        with open(os.path.join(self.test_root, 'tree', 'fulltext3.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy3')
        with open(os.path.join(self.test_root, 'tree', 'fulltext4.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy4')
        with open(os.path.join(self.test_root, 'tree', 'fulltext6.js'), 'w', encoding='UTF-8') as fh:
            fh.write('dummy6')

        book = Book(Host(self.test_root))
        book.fulltext = {
            "20200101000000000": {
                'index.html': {
                    'content': 'dummy text 1 中文',
                    },
                'frame.html': {
                    'content': 'frame page content',
                    },
                },
            "20200101000000001": {
                'index.html': {
                    'content': 'dummy text 2 中文',
                    },
                },
            }

        book.save_fulltext_files()

        with open(os.path.join(self.test_root, 'tree', 'fulltext.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.fulltext({
 "20200101000000000": {
  "index.html": {
   "content": "dummy text 1 中文"
  },
  "frame.html": {
   "content": "frame page content"
  }
 }
})""")
        with open(os.path.join(self.test_root, 'tree', 'fulltext1.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), """/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.fulltext({
 "20200101000000001": {
  "index.html": {
   "content": "dummy text 2 中文"
  }
 }
})""")
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'fulltext2.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'fulltext3.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'fulltext4.js')))
        self.assertFalse(os.path.exists(os.path.join(self.test_root, 'tree', 'fulltext5.js')))
        self.assertTrue(os.path.exists(os.path.join(self.test_root, 'tree', 'fulltext6.js')))

    def test_init_backup(self):
        book = Book(Host(self.test_root))

        book.init_backup(True)
        self.assertRegex(
            book.backup_dir,
            r'^' + re.escape(os.path.join(self.test_root, WSB_DIR, 'backup', '')) + r'\d{17}$',
            )

        ts = util.datetime_to_id()
        book.init_backup(ts)
        self.assertEqual(
            book.backup_dir,
            os.path.join(self.test_root, WSB_DIR, 'backup', ts),
            )

        book.init_backup(False)
        self.assertIsNone(book.backup_dir)

    def test_backup01(self):
        """A common case."""
        test_file = os.path.join(self.test_root, 'tree', 'meta.js')
        os.makedirs(os.path.dirname(test_file))
        with open(test_file, 'w', encoding='UTF-8') as fh:
            fh.write('abc')

        book = Book(Host(self.test_root))
        book.init_backup()
        book.backup(test_file)

        with open(os.path.join(book.backup_dir, 'tree', 'meta.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'abc')

    def test_backup02(self):
        """A common directory case."""
        test_dir = os.path.join(self.test_root, 'tree')
        os.makedirs(test_dir)
        with open(os.path.join(test_dir, 'meta.js'), 'w', encoding='UTF-8') as fh:
            fh.write('abc')
        with open(os.path.join(test_dir, 'toc.js'), 'w', encoding='UTF-8') as fh:
            fh.write('def')

        book = Book(Host(self.test_root))
        book.init_backup()
        book.backup(test_dir)

        with open(os.path.join(book.backup_dir, 'tree', 'meta.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'abc')
        with open(os.path.join(book.backup_dir, 'tree', 'toc.js'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'def')

    def test_backup03(self):
        """Pass if backup_dir not set."""
        test_file = os.path.join(self.test_wsbdir, 'icon', 'test.txt')
        os.makedirs(os.path.dirname(test_file))
        with open(test_file, 'w', encoding='UTF-8') as fh:
            fh.write('abc')
        test_stat = os.stat(test_file)

        book = Book(Host(self.test_root))
        book.backup(test_file)

        self.assertListEqual(os.listdir(self.test_wsbdir), ['icon'])

    @mock.patch('webscrapbook.scrapbook.host.FileLock')
    def test_get_lock01(self, mock_filelock):
        self.create_general_config()
        host = Host(self.test_root)
        book = Book(host)
        book.get_lock('test')
        mock_filelock.assert_called_once_with(host, 'book--test')

    @mock.patch('webscrapbook.scrapbook.host.FileLock')
    def test_get_lock02(self, mock_filelock):
        """With parameters"""
        self.create_general_config()
        host = Host(self.test_root)
        book = Book(host)
        book.get_lock('test',
            timeout=10, stale=120, poll_interval=0.3, assume_acquired=True)
        mock_filelock.assert_called_once_with(host, 'book--test',
            timeout=10, stale=120, poll_interval=0.3, assume_acquired=True)

    @mock.patch('webscrapbook.scrapbook.book.Book.get_lock')
    def test_get_tree_lock01(self, mock_get_lock):
        self.create_general_config()
        host = Host(self.test_root)
        book = Book(host)
        book.get_tree_lock()
        mock_get_lock.assert_called_once_with('tree')

    @mock.patch('webscrapbook.scrapbook.book.Book.get_lock')
    def test_get_tree_lock02(self, mock_get_lock):
        """With parameters"""
        self.create_general_config()
        host = Host(self.test_root)
        book = Book(host)
        book.get_tree_lock(timeout=10, stale=120, poll_interval=0.3, assume_acquired=True)
        mock_get_lock.assert_called_once_with('tree',
            timeout=10, stale=120, poll_interval=0.3, assume_acquired=True)

    def test_get_index_paths01(self):
        self.create_general_config()
        book = Book(Host(self.test_root))
        self.assertEqual(book.get_index_paths('20200101000000000/index.html'), ['index.html'])
        self.assertEqual(book.get_index_paths('20200101000000000.html'), ['20200101000000000.html'])
        self.assertEqual(book.get_index_paths('20200101000000000.htz'), ['index.html'])

    def test_get_index_paths02(self):
        """MAFF with single page"""
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'data'))
        archive_file = os.path.join(self.test_root, 'data', '20200101000000000.maff')
        with zipfile.ZipFile(archive_file, 'w') as zh:
            zh.writestr('20200101000000000/index.html', """dummy""")
        book = Book(Host(self.test_root))

        self.assertEqual(book.get_index_paths('20200101000000000.maff'), ['20200101000000000/index.html'])

    def test_get_index_paths03(self):
        """MAFF with multiple pages"""
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'data'))
        archive_file = os.path.join(self.test_root, 'data', '20200101000000000.maff')
        with zipfile.ZipFile(archive_file, 'w') as zh:
            zh.writestr('20200101000000000/index.html', """dummy""")
            zh.writestr('20200101000000001/index.html', """dummy""")
        book = Book(Host(self.test_root))

        self.assertEqual(book.get_index_paths('20200101000000000.maff'), ['20200101000000000/index.html', '20200101000000001/index.html'])

    def test_get_index_paths04(self):
        """MAFF with no page"""
        self.create_general_config()
        os.makedirs(os.path.join(self.test_root, 'data'))
        archive_file = os.path.join(self.test_root, 'data', '20200101000000000.maff')
        with zipfile.ZipFile(archive_file, 'w') as zh:
            pass
        book = Book(Host(self.test_root))

        self.assertEqual(book.get_index_paths('20200101000000000.maff'), [])

    def test_get_icon_file01(self):
        """Pass if file not exist."""
        book = Book(Host(self.test_root))

        self.assertIsNone(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': 'http://example.com',
            }))

        self.assertIsNone(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': 'data:image/bmp;base64,Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA',
            }))

        self.assertIsNone(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': '//example.com',
            }))

        self.assertIsNone(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': '/favicon.ico',
            }))

        self.assertIsNone(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': '',
            }))

        self.assertIsNone(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': '?id=123',
            }))

        self.assertIsNone(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': '#test',
            }))

        self.assertEqual(book.get_icon_file({
            'icon': 'favicon.ico?id=123#test',
            }),
            os.path.join(book.data_dir, 'favicon.ico'),
            )

        self.assertEqual(book.get_icon_file({
            'icon': '%E4%B8%AD%E6%96%87%231.ico?id=123#test',
            }),
            os.path.join(book.data_dir, '中文#1.ico'),
            )

        self.assertEqual(book.get_icon_file({
            'index': '20200101000000000/index.html',
            'icon': 'favicon.ico?id=123#test',
            }),
            os.path.join(book.data_dir, '20200101000000000', 'favicon.ico'),
            )

        self.assertEqual(book.get_icon_file({
            'index': '20200101000000000.html',
            'icon': 'favicon.ico?id=123#test',
            }),
            os.path.join(book.data_dir, 'favicon.ico'),
            )

        self.assertEqual(book.get_icon_file({
            'index': '20200101000000000.maff',
            'icon': 'favicon.ico?id=123#test',
            }),
            os.path.join(book.data_dir, 'favicon.ico'),
            )

        self.assertEqual(book.get_icon_file({
            'index': '20200101000000000.maff',
            'icon': '.wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp?id=123#test',
            }),
            os.path.join(book.tree_dir, 'favicon', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp'),
            )

    def test_load_note_file01(self):
        """Test for common note file wrapper."""
        test_file = os.path.join(self.test_root, 'index.html')
        with open(test_file, 'w', encoding='UTF-8') as f:
            f.write("""\
<!DOCTYPE html><html><head><meta charset="UTF-8"><meta name="viewport" content="width=device-width"><style>pre{white-space: pre-wrap; overflow-wrap: break-word;}</style></head><body><pre>
Note content
2nd line
3rd line
</pre></body></html>""")

        book = Book(Host(self.test_root))
        content = book.load_note_file(test_file)
        self.assertEqual(content, """\
Note content
2nd line
3rd line""")

    def test_load_note_file02(self):
        """Test for common legacy note file wrapper."""
        test_file = os.path.join(self.test_root, 'index.html')
        with open(test_file, 'w', encoding='UTF-8') as f:
            f.write("""\
<html><head><meta http-equiv="Content-Type" content="text/html;Charset=UTF-8"></head><body><pre>
Note content
2nd line
3rd line
</pre></body></html>""")

        book = Book(Host(self.test_root))
        content = book.load_note_file(test_file)
        self.assertEqual(content, """\
Note content
2nd line
3rd line""")

    def test_load_note_file03(self):
        """Return original text if malformatted."""
        test_file = os.path.join(self.test_root, 'index.html')
        html = """\
<html><head><meta http-equiv="Content-Type" content="text/html;Charset=UTF-8"></head><body>
Note content
2nd line
3rd line
</body></html>"""
        with open(test_file, 'w', encoding='UTF-8') as f:
            f.write(html)

        book = Book(Host(self.test_root))
        content = book.load_note_file(test_file)
        self.assertEqual(content, html)

    def test_save_note_file01(self):
        """Test saving. Enforce LF linefeeds."""
        test_file = os.path.join(self.test_root, 'index.html')

        book = Book(Host(self.test_root))
        book.save_note_file(test_file, """\
Note content
2nd line
3rd line""")

        with open(test_file, encoding='UTF-8', newline='') as fh:
            self.assertEqual(fh.read(), """\
<!DOCTYPE html><html><head>\
<meta charset="UTF-8">\
<meta name="viewport" content="width=device-width">\
<style>pre { white-space: pre-wrap; overflow-wrap: break-word; }</style>\
</head><body><pre>
Note content
2nd line
3rd line
</pre></body></html>""")

if __name__ == '__main__':
    unittest.main()



bidi_dir = 'ltr'


#########################################################################
# scrapbook.cache
#########################################################################

# map.html
cache_index_toggle_all = 'Toggle all'
cache_index_search_link_title = 'Search'
cache_index_source_link_title = 'Source link'

# search.html
cache_search_title = '%book% :: Search'
cache_search_view_in_map = 'View in map'
cache_search_start = 'go'
cache_search_help_label = 'Search syntax help'
cache_search_help_desc = """\
• Input keywords search from title, content, or comment.
• Use space to separate multiple keywords. For example, “w3c organization” means items containing “w3c” and “organization”.
• Use double quotes to demark a complete phrase, a literal double quote can be escaped by doubling. For example, “"Tom ""loves"" Mary."” means items containing “Tom "loves" Mary.”
• Use minus sign to exclude a keyword. For example, “-virus” means items without “virus”.
• Use “<command>:<keyword>” to specify a special search condition. A minus sign can be prefixed for exclusion or reversion. Available commands include:
  • mc: each subsequent keyword matches case-sensitively. For example, “mc: CIA FBI -mc: president” means items containing “CIA” and “FBI” case-sensitively, and “president” case-insensitively.
  • re: each subsequent keyword is treated as a regular expression. For example, “re: \\bcolou?r\\b -re: 1+1=2” means items that match regular expression “\\bcolou?r\\b” and contain keyword “1+1=2”.
  • id: item whose ID equal to the keyword (or match the regular expression). Multiple values are “or”-connected. For example, “id:2020 id:2021” means items of ID “2020” or “2021”; “-id:2020” means items whose ID is not “2020”.
  • type: items whose type equal to the keyword (or match the regular expression). Multiple values are “or”-connected. Available types are “” (page), “bookmark”, “file”, “note”, etc. For example, “type: type:bookmark” means items whose type is page or bookmark.
  • title: items whose title contains the keyword.
  • content: items whose fulltext contains the keyword.
  • comment: items whose comment contains the keyword.
  • tcc: items whose title, fulltext, or comment contains the keyword.
  • index: items whose index file path contains the keyword.
  • source: items whose source URL contains the keyword.
  • icon: items whose icon URL contains the keyword.
  • create: items whose create time matches the condition. The time condition is an interval with 0-17 digits, followed by a minus sign optionally, and then followed by 0-17 digits. The two 17-digit numbers means the year (4 digits), month (01-12), day (01-31), hours (00-59), minutes (00-59), seconds (00-59), and milliseconds (000-999) in local datetime. Each omitted digit is assumed to be a “0”, except that “999...” is assumed if the end datetime is totally omitted. For example, “create:2014-2015” means since 2014 until 2015; “create:-201309” means before Sep 2013; and “create:20110825” means after Aug 25, 2011.
  • modify: items whose modify time matches the condition. Time format is same as create.
  • marked: marked items.
  • locked: locked items.
  • file: items whose filename contains the keyword.
  • root: items under the item of ID. Multiple values are “or”-connected.
  • book: items in the specific scrapbook (by name). Multiple values are “or”-connected.
  • sort: sort search results using the specific condition, which can be id, title, comment, content, source, type, create, or modify. For example, “sort:id -sort:modify” means sorting by ID in acending order and then sorting by modify time in descending order."""
cache_search_result = 'Found %length% results:'
cache_search_result_named = '(%name%) Found %length% results:'
cache_search_sort_last_created = 'Last Created'
cache_search_sort_last_modified = 'Last Modified'
cache_search_sort_title = 'Sort by title'
cache_search_sort_id = 'Sort by ID'
